r"""
===========
Simulations
===========
This module contains classes that perform simulations directly on the images
to obtain structural information

.. autoclass:: RandomWalk
    :members:

.. autoclass:: Porosimetry
    :members:

"""
from .__RandomWalk__ import RandomWalk
from .__Porosimetry__ import Porosimetry



import porespy as ps
import scipy as sp
import git


class PackageTest():
    def setup_class(self):
        pass

    def test_version_number_and_git_tag_agree(self):
        repo = git.Repo(search_parent_directories=True)
        tag = repo.git.describe("--tags")
        tag = tag.strip('vV')  # Remove 'v' or 'V' from tag if present
        tag = tag.split('-')[0]  # Remove hash from tag number if present
        assert ps.__version__ == tag



# ***********************************************************************************
# Copyright 2010 - 2016 Paulo A. Herrera. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2.Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY PAULO A. HERRERA ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING
# BUT NOT LIMITED TO, PROCUREMEN OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ***********************************************************************************

# **************************************
# *  Low level Python library to       *
# *  export data to binary VTK file.   *
# **************************************

from .base import writeBlockSize, writeArrayToFile, writeArraysToFile
from .xml import XmlWriter
import sys
import os

# ================================
#            VTK Types
# ================================

#     FILE TYPES


class VtkFileType:

    def __init__(self, name, ext):
        self.name = name
        self.ext = ext

    def __str__(self):
        return "Name: %s  Ext: %s \n" % (self.name, self.ext)

VtkImageData = VtkFileType("ImageData", ".vti")
VtkPolyData = VtkFileType("PolyData", ".vtp")
VtkRectilinearGrid = VtkFileType("RectilinearGrid", ".vtr")
VtkStructuredGrid = VtkFileType("StructuredGrid", ".vts")
VtkUnstructuredGrid = VtkFileType("UnstructuredGrid", ".vtu")

#    DATA TYPES


class VtkDataType:

    def __init__(self, size, name):
        self.size = size
        self.name = name

    def __str__(self):
        return "Type: %s  Size: %d \n" % (self.name, self.size)

VtkInt8 = VtkDataType(1, "Int8")
VtkUInt8 = VtkDataType(1, "UInt8")
VtkInt16 = VtkDataType(2, "Int16")
VtkUInt16 = VtkDataType(2, "UInt16")
VtkInt32 = VtkDataType(4, "Int32")
VtkUInt32 = VtkDataType(4, "UInt32")
VtkInt64 = VtkDataType(8, "Int64")
VtkUInt64 = VtkDataType(8, "UInt64")
VtkFloat32 = VtkDataType(4, "Float32")
VtkFloat64 = VtkDataType(8, "Float64")

# Map numpy to VTK data types
np_to_vtk = {'int8': VtkInt8,
             'uint8': VtkUInt8,
             'int16': VtkInt16,
             'uint16': VtkUInt16,
             'int32': VtkInt32,
             'uint32': VtkUInt32,
             'int64': VtkInt64,
             'uint64': VtkUInt64,
             'float32': VtkFloat32,
             'float64': VtkFloat64}

#    CELL TYPES


class VtkCellType:

    def __init__(self, tid, name):
        self.tid = tid
        self.name = name

    def __str__(self):
        return "VtkCellType( %s ) \n" % (self.name)

VtkVertex = VtkCellType(1, "Vertex")
VtkPolyVertex = VtkCellType(2, "PolyVertex")
VtkLine = VtkCellType(3, "Line")
VtkPolyLine = VtkCellType(4, "PolyLine")
VtkTriangle = VtkCellType(5, "Triangle")
VtkTriangleStrip = VtkCellType(6, "TriangleStrip")
VtkPolygon = VtkCellType(7, "Polygon")
VtkPixel = VtkCellType(8, "Pixel")
VtkQuad = VtkCellType(9, "Quad")
VtkTetra = VtkCellType(10, "Tetra")
VtkVoxel = VtkCellType(11, "Voxel")
VtkHexahedron = VtkCellType(12, "Hexahedron")
VtkWedge = VtkCellType(13, "Wedge")
VtkPyramid = VtkCellType(14, "Pyramid")
VtkQuadraticEdge = VtkCellType(21, "Quadratic_Edge")
VtkQuadraticTriangle = VtkCellType(22, "Quadratic_Triangle")
VtkQuadraticQuad = VtkCellType(23, "Quadratic_Quad")
VtkQuadraticTetra = VtkCellType(24, "Quadratic_Tetra")
VtkQuadraticHexahedron = VtkCellType(25, "Quadratic_Hexahedron")

# ==============================
#       Helper functions
# ==============================


def _mix_extents(start, end):
    assert (len(start) == len(end) == 3)
    string = "%d %d %d %d %d %d" % (start[0], end[0], start[1], end[1],
                                    start[2], end[2])
    return string


def _array_to_string(a):
    s = "".join([repr(num) + " " for num in a])
    return s


def _get_byte_order():
    if sys.byteorder == "little":
        return "LittleEndian"
    else:
        return "BigEndian"

# ================================
#        VtkGroup class
# ================================


class VtkGroup:

    def __init__(self, filepath):
        """ Creates a VtkGroup file that is stored in filepath.

            PARAMETERS:
                filepath: filename without extension.
        """
        self.xml = XmlWriter(filepath + ".pvd")
        self.xml.openElement("VTKFile")
        self.xml.addAttributes(type="Collection", version="0.1",
                               byte_order=_get_byte_order())
        self.xml.openElement("Collection")
        self.root = os.path.dirname(filepath)

    def save(self):
        """ Closes this VtkGroup. """
        self.xml.closeElement("Collection")
        self.xml.closeElement("VTKFile")
        self.xml.close()

    def addFile(self, filepath, sim_time):
        """ Adds file to this VTK group.

            PARAMETERS:
                filepath: full path to VTK file.
                sim_time: simulated time.
        """
        # TODO: Check what the other attributes are for.
        filename = os.path.relpath(filepath, start=self.root)
        self.xml.openElement("DataSet")
        self.xml.addAttributes(timestep=sim_time, group="",
                               part="0", file=filename)
        self.xml.closeElement()

# ================================
#        VtkFile class
# ================================


class VtkFile:

    def __init__(self, filepath, ftype, largeFile=False):
        """
    PARAMETERS:
        filepath: filename without extension.
        ftype: file type, e.g. VtkImageData, etc.
        largeFile: If size of the stored data cannot be represented by a UInt32
        """
        self.ftype = ftype
        self.filename = filepath + ftype.ext
        self.xml = XmlWriter(self.filename)
        self.offset = 0  # offset in bytes after beginning of binary section
        self.appendedDataIsOpen = False
        self.xml.openElement("VTKFile").addAttributes(type=ftype.name,
                                                      version="1.0",
                                                      byte_order=_get_byte_order(),
                                                      header_type="UInt64")

    def getFileName(self):
        """ Returns absolute path to this file. """
        return os.path.abspath(self.filename)

    def openPiece(self, start=None, end=None,
                  npoints=None, ncells=None,
                  nverts=None, nlines=None, nstrips=None, npolys=None):
        """ Open piece section.

        PARAMETERS:
            Next two parameters must be given together.
            start: array or list with start indexes in each direction.
            end:   array or list with end indexes in each direction.

            npoints: number of points in piece (int).
            ncells: number of cells in piece (int). If present,
                    npoints must also be given.

            All the following parameters must be given together with npoints.
            They should all be integer values.
            nverts: number of vertices.
            nlines: number of lines.
            nstrips: number of strips.
            npolys: number of .

        RETURNS:
            this VtkFile to allow chained calls.
        """
        # TODO: Check what are the requirements for each type of grid.

        self.xml.openElement("Piece")
        if (start and end):
            ext = _mix_extents(start, end)
            self.xml.addAttributes(Extent=ext)

        elif (ncells and npoints):
            self.xml.addAttributes(NumberOfPoints=npoints, NumberOfCells=ncells)

        elif npoints or nverts or nlines or nstrips or npolys:
            if npoints is None:
                npoints = str(0)
            if nverts is None:
                nverts = str(0)
            if nlines is None:
                nlines = str(0)
            if nstrips is None:
                nstrips = str(0)
            if npolys is None:
                npolys = str(0)
            self.xml.addAttributes(NumberOfPoints=npoints,
                                   NumberOfVerts=nverts,
                                   NumberOfLines=nlines,
                                   NumberOfStrips=nstrips,
                                   NumberOfPolys=npolys)
        else:
            assert(False)

        return self

    def closePiece(self):
        self.xml.closeElement("Piece")

    def openData(self, nodeType, scalars=None, vectors=None,
                 normals=None, tensors=None, tcoords=None):
        """ Open data section.

            PARAMETERS:
                nodeType: Point or Cell.
                scalars: default data array name for scalar data.
                vectors: default data array name for vector data.
                normals: default data array name for normals data.
                tensors: default data array name for tensors data.
                tcoords: dafault data array name for tcoords data.

            RETURNS:
                this VtkFile to allow chained calls.
        """
        self.xml.openElement(nodeType + "Data")
        if scalars:
            self.xml.addAttributes(scalars=scalars)
        if vectors:
            self.xml.addAttributes(vectors=vectors)
        if normals:
            self.xml.addAttributes(normals=normals)
        if tensors:
            self.xml.addAttributes(tensors=tensors)
        if tcoords:
            self.xml.addAttributes(tcoords=tcoords)

        return self

    def closeData(self, nodeType):
        """ Close data section.

            PARAMETERS:
                nodeType: Point or Cell.

            RETURNS:
                this VtkFile to allow chained calls.
        """
        self.xml.closeElement(nodeType + "Data")

    def openGrid(self, start=None, end=None, origin=None, spacing=None):
        """ Open grid section.

    PARAMETERS:
        start: array or list of start indexes. Required for Structured,
            Rectilinear and ImageData grids.
        end: array or list of end indexes. Required for Structured,
            Rectilinear and ImageData grids.
        origin: 3D array or list with grid origin.
            Only required for ImageData grids.
        spacing: 3D array or list with grid spacing.
            Only required for ImageData grids.

    RETURNS:
        this VtkFile to allow chained calls.
        """
        gType = self.ftype.name
        self.xml.openElement(gType)
        if (gType == VtkImageData.name):
            if (not start or not end or not origin or not spacing):
                assert(False)
            ext = _mix_extents(start, end)
            self.xml.addAttributes(WholeExtent=ext,
                                   Origin=_array_to_string(origin),
                                   Spacing=_array_to_string(spacing))

        elif (gType == VtkStructuredGrid.name or
              gType == VtkRectilinearGrid.name):
            if (not start or not end):
                assert (False)
            ext = _mix_extents(start, end)
            self.xml.addAttributes(WholeExtent=ext)

        return self

    def closeGrid(self):
        """ Closes grid element.

            RETURNS:
                this VtkFile to allow chained calls.
        """
        self.xml.closeElement(self.ftype.name)

    def addHeader(self, name, dtype, nelem, ncomp):
        """ Adds data array description to xml header section.

            PARAMETERS:
                name: data array name.
                dtype: string describing type of the data.
                       Format is the same as used by numpy, e.g. 'float64'.
                nelem: number of elements in the array.
                ncomp: number of components, 1 (=scalar) and 3 (=vector).

            RETURNS:
                This VtkFile to allow chained calls.

            NOTE: This is a low level function. Use addData if you want
                  to add a numpy array.
        """
        dtype = np_to_vtk[dtype]

        self.xml.openElement("DataArray")
        self.xml.addAttributes(Name=name,
                               NumberOfComponents=ncomp,
                               type=dtype.name,
                               format="appended",
                               offset=self.offset)
        self.xml.closeElement()

        self.offset += nelem * ncomp * dtype.size + 8
        # add 8 to indicate array size
        return self

    def addData(self, name, data):
        """ Adds array description to xml header section.

             PARAMETERS:
                name: data array name.
                data: one numpy array or a tuple with 3 numpy arrays.
                    If a tuple, the individual
                    arrays must represent the components of a vector field.
                    All arrays must be one dimensional or three-dimensional.
        """
        if type(data).__name__ == "tuple":  # vector data
            assert (len(data) == 3)
            x = data[0]
            self.addHeader(name, x.dtype.name, x.size, 3)
        elif type(data).__name__ == "ndarray":
            if data.ndim == 1 or data.ndim == 3:
                self.addHeader(name, data.dtype.name, data.size, 1)
            else:
                assert False, "Bad array shape: " + str(data.shape)
        else:
            assert False, "Argument must be a Numpy array"

    def appendHeader(self, dtype, nelem, ncomp):
        """ This function only writes the size of the data block
            that will be appended. The data itself must be written immediately
            after calling this function.

            PARAMETERS:
                dtype: string with data type representation (same as numpy).
                    For example, 'float64'
                nelem: number of elements.
                ncomp: number of components, 1 (=scalar) or 3 (=vector).
        """
        self.openAppendedData()
        dsize = np_to_vtk[dtype].size
        block_size = dsize * ncomp * nelem
        if self.largeFile is False:
            writeBlockSize(self.xml.stream, block_size)
        else:
            writeBlockSize64Bit(self.xml.stream, block_size)

    def appendData(self, data):
        """ Append data to binary section.
    This function writes the header section and the data to the binary file.

    PARAMETERS:
        data: one numpy array or a tuple with 3 numpy arrays.
              If a tuple, the individual arrays must represent
              the components of a vector field.
              All arrays must be one dimensional or three-dimensional.
              The order of the arrays must coincide with the numbering
              scheme of the grid.

    RETURNS:
        This VtkFile to allow chained calls

    TODO: Extend this function to accept contiguous C order arrays.
"""
        self.openAppendedData()

        if type(data).__name__ == 'tuple':  # 3 numpy arrays
            ncomp = len(data)
            assert (ncomp == 3)
            dsize = data[0].dtype.itemsize
            nelem = data[0].size
            block_size = ncomp * nelem * dsize
            writeBlockSize(self.xml.stream, block_size)
            x, y, z = data[0], data[1], data[2]
            writeArraysToFile(self.xml.stream, x, y, z)
        elif type(data).__name__ == 'ndarray' and (data.ndim == 1 or data.ndim == 3):
            # single numpy array
            ncomp = 1
            dsize = data.dtype.itemsize
            nelem = data.size
            block_size = ncomp * nelem * dsize
            writeBlockSize(self.xml.stream, block_size)
            writeArrayToFile(self.xml.stream, data)
        else:
            assert False

        return self

    def openAppendedData(self):
        """ Opens binary section.

            It is not necessary to explicitly call this function
            from an external library.
        """
        if not self.appendedDataIsOpen:
            self.xml.openElement("AppendedData").addAttributes(encoding="raw")
            self.xml.addText("_")
            self.appendedDataIsOpen = True

    def closeAppendedData(self):
        """ Closes binary section.

            It is not necessary to explicitly call this function
            from an external library.
        """
        self.xml.closeElement("AppendedData")

    def openElement(self, tagName):
        """ Useful to add elements such as:Coordinates, Points, Verts, etc. """
        self.xml.openElement(tagName)

    def closeElement(self, tagName):
        self.xml.closeElement(tagName)

    def save(self):
        """ Closes file """
        if self.appendedDataIsOpen:
            self.xml.closeElement("AppendedData")
        self.xml.closeElement("VTKFile")
        self.xml.close()



import scipy as sp
import scipy.ndimage as spim
from skimage.morphology import ball, disk, square, cube
from array_split import shape_split


def subdivide(im, divs=2):
    r"""
    Returns slices into an image describing the specified number of sub-arrays.
    This function is useful for performing operations on smaller images for
    memory or speed.  Note that most typical operations this will NOT work,
    since the image borders would cause artifacts (e.g. ``distance_transform``)

    Parameters
    ----------
    im : ND-array
        The image of the porous media

    divs : scalar or array_like
        The number of sub-divisions to create in each axis of the image.  If a
        scalar is given it is assume this value applies in all dimensions.

    Returns
    -------
    An ND-array containing slice objects for indexing into ``im`` that extract
    the sub-divided arrays.

    Notes
    -----
    This method uses the
    `array_split package <https://github.com/array-split/array_split>`_ which
    offers the same functionality as the ``split`` method of Numpy's ND-array,
    but supports the splitting multidimensional arrays in all dimensions.

    Examples
    --------
    >>> import porespy as ps
    >>> import matplotlib.pyplot as plt
    >>> im = ps.generators.blobs(shape=[200, 200])
    >>> s = ps.tools.subdivide(im, divs=[2, 2])

    ``s`` contains an array with the shape given by ``divs``.  To access the
    first and last quadrants of ``im`` use:
    >>> print(im[s[0, 0]].shape)
    (100, 100)
    >>> print(im[s[1, 1]].shape)
    (100, 100)

    It can be easier to index the array with the slices by applying ``flatten``
    first:
    >>> s_flat = s.flatten()
    >>> for i in s_flat:
    ...     print(im[i].shape)
    (100, 100)
    (100, 100)
    (100, 100)
    (100, 100)
    """
    # Expand scalar divs
    if sp.array(divs, ndmin=1).size == 1:
        divs = [divs for i in range(im.ndim)]
    s = shape_split(im.shape, axis=divs)
    return s


def get_slice(im, center, size, pad=0):
    r"""
    Given a ``center`` location and ``radius`` of a feature, returns the slice
    object into the ``im`` that bounds the feature but does not extend beyond
    the image boundaries.

    Parameters
    ----------
    im : ND-image
        The image of the porous media

    center : array_like
        The coordinates of the center of the feature of interest

    size : array_like or scalar
        The size of the feature in each direction.  If a scalar is supplied,
        this implies the same size in all directions.

    pad : scalar or array_like
        The amount to pad onto each side of the slice.  The default is 0.  A
        scalar value will increase the slice size equally in all directions,
        while an array the same shape as ``im.shape`` can be passed to pad
        a specified amount in each direction.

    Returns
    -------
    A list of slice objects, each indexing into one dimension of the image.
    """
    p = sp.ones(shape=im.ndim, dtype=int)*sp.array(pad)
    s = sp.ones(shape=im.ndim, dtype=int)*sp.array(size)
    slc = []
    for dim in range(im.ndim):
        lower_im = sp.amax((center[dim] - s[dim] - p[dim], 0))
        upper_im = sp.amin((center[dim] + s[dim] + 1 + p[dim], im.shape[dim]))
        slc.append(slice(lower_im, upper_im))
    return slc


def find_outer_region(im, r=0):
    r"""
    Finds regions of the image that are outside of the solid matrix.  This
    function uses the rolling ball method to define where the outer region
    ends and the void space begins.

    This function is particularly useful for samples that do not fill the
    entire rectangular image, such as cylindrical cores or samples with non-
    parallel faces.

    Parameters
    ----------
    im : ND-array
        Image of the porous material with 1's for void and 0's for solid

    r : scalar
        The radius of the rolling ball to use.  If not specified then a value
        is calculated as twice maximum of the distance transform.  The image
        size is padded by this amount in all directions, so the image can
        become quite large and unwieldy if too large a value is given.

    Returns
    -------
    A boolean mask the same shape as ``im``, containing True in all voxels
    identified as *outside* the sample.

    """
    if r == 0:
        dt = spim.distance_transform_edt(input=im)
        r = int(sp.amax(dt))*2
    im_padded = sp.pad(array=im, pad_width=r, mode='constant',
                       constant_values=True)
    dt = spim.distance_transform_edt(input=im_padded)
    seeds = (dt >= r) + get_border(shape=im_padded.shape)
    # Remove seeds not connected to edges
    labels = spim.label(seeds)[0]
    mask = labels == 1  # Assume label of 1 on edges, assured by adding border
    dt = spim.distance_transform_edt(~mask)
    outer_region = dt < r
    outer_region = extract_subsection(im=outer_region, shape=im.shape)
    return outer_region


def extract_cylinder(im, r=None, axis=0):
    r"""
    Returns a cylindrical section of the image of specified radius. This is
    useful for making square images look like cylindrical cores such as those
    obtained from X-ray tomography.

    Parameters
    ----------
    im : ND-array
        The image of the porous material

    r : scalr
        The radius of the cylinder to extract.  If none if given then the
        default is the largest cylinder that can fit inside the x-y plane.

    axis : scalar
        The axis along with the cylinder will be oriented.

    Returns
    -------
    An ND-image the same size ``im`` with True values indicating the void space
    but with the sample trimmed to a cylindrical section in the center of the
    image.  The region outside the cylindrical section is labeled with True
    values since it is open space.
    """
    if r is None:
        a = list(im.shape)
        a.pop(axis)
        r = sp.amin(a)/2
    dim = [range(int(-s/2), int(s/2)) for s in im.shape]
    inds = sp.meshgrid(*dim, indexing='ij')
    inds[axis] = inds[axis]*0
    d = sp.sqrt(sp.sum(sp.square(inds), axis=0))
    mask = d <= r
    im[~mask] = True
    return im


def extract_subsection(im, shape):
    r"""
    Extracts the middle section of a image

    Parameters
    ----------
    im : ND-array
        Image from which to extract the subsection

    shape : array_like
        Can either specify the size of the extracted section or the fractional
        size of the image to extact.

    Returns
    -------
    An ND-array of size given by the ``shape`` argument, taken from the center
    of the image.

    Examples
    --------
    >>> import scipy as sp
    >>> from porespy.tools import extract_subsection
    >>> im = sp.array([[1, 1, 1, 1], [1, 2, 2, 2], [1, 2, 3, 3], [1, 2, 3, 4]])
    >>> print(im)
    [[1 1 1 1]
     [1 2 2 2]
     [1 2 3 3]
     [1 2 3 4]]
    >>> im = extract_subsection(im=im, shape=[2, 2])
    >>> print(im)
    [[2 2]
     [2 3]]

    """
    # Check if shape was given as a fraction
    shape = sp.array(shape)
    if shape[0] < 1:
        shape = sp.array(im.shape)*shape
    center = sp.array(im.shape)/2
    s_im = []
    for dim in range(im.ndim):
        r = shape[dim]/2
        lower_im = sp.amax((center[dim]-r, 0))
        upper_im = sp.amin((center[dim]+r, im.shape[dim]))
        s_im.append(slice(int(lower_im), int(upper_im)))
    return im[s_im]


def extend_slice(s, shape, pad=1):
    r"""
    Adjust slice indices to include additional voxles around the slice.  The
    key to this function is that is does bounds checking to ensure the indices
    don't extend outside the image.

    Parameters
    ----------
    s : list of slice objects
         A list (or tuple) of N slice objects, where N is the number of
         dimensions in the image.

    shape : array_like
        The shape of the image into which the slice objects apply.  This is
        used to check the bounds to prevent indexing beyond the image.

    pad : int
        The number of voxels to expand in each direction.

    Returns
    -------
    A list slice objects with the start and stop attributes respectively
    incremented and decremented by 1, without extending beyond the image
    boundaries.

    Examples
    --------
    >>> from scipy.ndimage import label, find_objects
    >>> from porespy.tools import extend_slice
    >>> im = sp.array([[1, 0, 0], [1, 0, 0], [0, 0, 1]])
    >>> labels = label(im)[0]
    >>> s = find_objects(labels)

    Using the slices returned by ``find_objects``, set the first label to 3

    >>> labels[s[0]] = 3
    >>> print(labels)
    [[3 0 0]
     [3 0 0]
     [0 0 2]]

    Next extend the slice, and use it to set the values to 4

    >>> s_ext = extend_slice(s[0], shape=im.shape, pad=1)
    >>> labels[s_ext] = 4
    >>> print(labels)
    [[4 4 0]
     [4 4 0]
     [4 4 2]]

    As can be seen by the location of the 4s, the slice was extended by 1, and
    also handled the extension beyond the boundary correctly.
    """
    a = []
    for i, dim in zip(s, shape):
        start = 0
        stop = dim
        if i.start - pad >= 0:
            start = i.start - pad
        if i.stop + pad < dim:
            stop = i.stop + pad
        a.append(slice(start, stop, None))
    return a


def binary_opening_fast(im, r, dt=None):
    r"""
    This function uses a shortcut to perform a morphological opening that does
    not slow down with larger structuring elements.  Because of the shortcut,
    it only applies to spherical structuring elements.

    Parameters
    ----------
    im : ND-array
        The image of the porous material with True values (or 1's) indicating
        the pore phase.

    dt : ND-array
        The distance transform of the pore space.  If none is provided, it will
        be calculated; however, providing one is a good idea since it will cut
        the processing time in half.

    r : scalar, int
        The radius of the spherical structuring element to apply

    Returns
    -------
    A binary image with ``True`` values in all locations where a sphere of size
    ``r`` could fit entirely within the pore space.

    """
    if dt is None:
        dt = spim.distance_transform_edt(im)
    seeds = dt > r
    im_opened = spim.distance_transform_edt(~seeds) <= r
    return im_opened


def randomize_colors(im, keep_vals=[0]):
    r'''
    Takes a greyscale image and randomly shuffles the greyscale values, so that
    all voxels labeled X will be labelled Y, and all voxels labeled Y will be
    labeled Z, where X, Y, Z and so on are randomly selected from the values
    in the input image.

    This function is useful for improving the visibility of images with
    neighboring regions that are only incrementally different from each other,
    such as that returned by `scipy.ndimage.label`.

    Parameters
    ----------
    im : array_like
        An ND image of greyscale values.

    keep_vals : array_like
        Indicate which voxel values should NOT be altered.  The default is
        `[0]` which is useful for leaving the background of the image
        untouched.

    Returns
    -------
    An image the same size and type as `im` but with the greyscale values
    reassigned.  The unique values in both the input and output images will
    be identical.

    Notes
    -----
    If the greyscale values in the input image are not contiguous then the
    neither will they be in the output.

    Examples
    --------
    >>> import porespy as ps
    >>> import scipy as sp
    >>> sp.random.seed(0)
    >>> im = sp.random.randint(low=0, high=5, size=[4, 4])
    >>> print(im)
    [[4 0 3 3]
     [3 1 3 2]
     [4 0 0 4]
     [2 1 0 1]]
    >>> im_rand = ps.tools.randomize_colors(im)
    >>> print(im_rand)
    [[2 0 4 4]
     [4 1 4 3]
     [2 0 0 2]
     [3 1 0 1]]

    As can be seen, the 2's have become 3, 3's have become 4, and 4's have
    become 2.  1's remained 1 by random accident.  0's remain zeros by default,
    but this can be controlled using the `keep_vals` argument.

    '''
    im_flat = im.flatten()
    keep_vals = sp.array(keep_vals)
    swap_vals = ~sp.in1d(im_flat, keep_vals)
    im_vals = sp.unique(im_flat[swap_vals])
    new_vals = sp.random.permutation(im_vals)
    im_map = sp.zeros(shape=[sp.amax(im_vals) + 1, ], dtype=int)
    im_map[im_vals] = new_vals
    im_new = im_map[im_flat]
    im_new = sp.reshape(im_new, newshape=sp.shape(im))
    return im_new


def make_contiguous(im):
    r"""
    Take an image with arbitrary greyscale values and adjust them to ensure
    all values fall in a contiguous range starting at 0.

    Parameters
    ----------
    im : array_like
        An ND array containing greyscale values

    Returns
    -------
    An ND-array the same size as ``im`` but with all values in contiguous
    orders.

    Example
    -------
    >>> import porespy as ps
    >>> import scipy as sp
    >>> im = sp.array([[0, 2, 9], [6, 8, 3]])
    >>> im = ps.tools.make_contiguous(im)
    >>> print(im)
    [[0 1 5]
     [3 4 2]]

    """
    im_flat = im.flatten()
    im_vals = sp.unique(im_flat)
    im_map = sp.zeros(shape=sp.amax(im_flat)+1)
    im_map[im_vals] = sp.arange(0, sp.size(sp.unique(im_flat)))
    im_new = im_map[im_flat]
    im_new = sp.reshape(im_new, newshape=sp.shape(im))
    im_new = sp.array(im_new, dtype=im_flat.dtype)
    return im_new


def get_border(shape, thickness=1, mode='edges'):
    r"""
    Creates an array of specified size with corners, edges or faces labelled as
    True.  This can be used as mask to manipulate values laying on the
    perimeter of an image.

    Parameters
    ----------
    shape : array_like
        The shape of the array to return.  Can be either 2D or 3D.

    thickness : scalar (default is 1)
        The number of pixels/voxels to place along perimeter.

    mode : string
        The type of border to create.  Options are 'faces', 'edges' (default)
        and 'corners'.  In 2D 'faces' and 'edges' give the same result.

    Returns
    -------
    An ND-array of specified shape with True values at the perimeter and False
    elsewhere.

    Examples
    --------
    >>> import porespy as ps
    >>> import scipy as sp
    >>> mask = ps.tools.get_border(shape=[3, 3], mode='corners')
    >>> print(mask)
    [[ True False  True]
     [False False False]
     [ True False  True]]
    >>> mask = ps.tools.get_border(shape=[3, 3], mode='edges')
    >>> print(mask)
    [[ True  True  True]
     [ True False  True]
     [ True  True  True]]
    """
    ndims = len(shape)
    t = thickness
    border = sp.ones(shape, dtype=bool)
    if mode == 'faces':
        if ndims == 2:
            border[t:-t, t:-t] = False
        if ndims == 3:
            border[t:-t, t:-t, t:-t] = False
    elif mode == 'edges':
        if ndims == 2:
            border[t:-t, t:-t] = False
        if ndims == 3:
            border[0::, t:-t, t:-t] = False
            border[t:-t, 0::, t:-t] = False
            border[t:-t, t:-t, 0::] = False
    elif mode == 'corners':
        if ndims == 2:
            border[t:-t, 0::] = False
            border[0::, t:-t] = False
        if ndims == 3:
            border[t:-t, 0::, 0::] = False
            border[0::, t:-t, 0::] = False
            border[0::, 0::, t:-t] = False
    return border



import scipy as sp
import scipy.ndimage as spim
from skimage.morphology import reconstruction, watershed
from porespy.tools import randomize_colors


def align_image_with_openpnm(im):
    r"""
    Rotates an image to agree with the coordinates used in OpenPNM.  It is
    unclear why they are not in agreement to start with.  This is necessary
    for overlaying the image and the network in Paraview.

    Parameters
    ----------
    im : ND-array
        The image to be rotated.  Can the Boolean image of the pore space or
        any other image of interest.

    Returns
    -------
    Returns the image rotated accordingly.
    """
    if im.ndim == 2:
        im = (sp.swapaxes(im, 1, 0))
        im = im[-1::-1, :]
    elif im.ndim == 3:
        im = (sp.swapaxes(im, 2, 0))
        im = im[:, -1::-1, :]
    return im


def partition_pore_space(im, peaks):
    r"""
    Applies a watershed segmentation to partition the distance transform into
    discrete pores.

    Parameters
    ----------
    im : ND-array
        Either the Boolean array of the pore space or a distance tranform.
        Passing in a pre-existing distance transform saves time, since one is
        calculated by the function if a Boolean array is passed.

    peaks : ND-array
        An array the same size as ``dt`` indicating where the pore centers are.
        If boolean, the peaks are labeled; if numeric then it's assumed that
        peaks have already been given desired labels.

    Returns
    -------
    A ND-array the same size as ``dt`` with regions belonging to each peak
    labelled.  The region number is randomized so that neighboring regions
    are contrasting colors for easier visualization.

    Notes
    -----
    Find the appropriate ``peaks`` array is the tricky part.  In principle,
    each local maxima in the distance transform is a pore center, but due to
    loss of fidelity in the voxel representation this leads to many extraneous
    peaks.  Several functions are available to help select peaks which can then
    be passed to this function.

    See Also
    --------
    snow

    """
    print('_'*60)
    print('Partitioning Pore Space using Marker Based Watershed')
    if im.dtype == bool:
        print('Boolean image received, applying distance transform')
        im = spim.distance_transform_edt(im)
    if peaks.dtype == bool:
        print('Boolean peaks received, applying labeling')
        peaks = spim.label(input=peaks)[0]
    regions = watershed(-im, markers=peaks)
    return regions



r'''
=============
Visualization
=============

This module contains functions for quickly visualizing 3D images in 2D views.

.. autofunction:: sem
.. autofunction:: xray
.. autofunction:: set_mpl_style


'''


from .__views__ import sem
from .__views__ import xray
from .__funcs__ import set_mpl_style



import scipy as sp
import scipy.spatial as sptl
import scipy.ndimage as spim
from skimage.segmentation import find_boundaries
from skimage.morphology import ball, disk, square, cube
from tqdm import tqdm


def insert_shape(im, center, element, value=1):
    r"""
    """
    if im.ndim != element.ndim:
        raise Exception('Image shape ' + str(im.shape) +
                        ' and element shape ' + str(element.shape) +
                        ' do not match')
    s_im = []
    s_el = []
    for dim in range(im.ndim):
        r = int(element.shape[dim]/2)
        lower_im = sp.amax((center[dim] - r, 0))
        upper_im = sp.amin((center[dim] + r + 1, im.shape[dim]))
        s_im.append(slice(lower_im, upper_im))
        lower_el = sp.amax((lower_im - center[dim] + r, 0))
        upper_el = sp.amin((upper_im - center[dim] + r, element.shape[dim]))
        s_el.append(slice(lower_el, upper_el))
    im[s_im] = im[s_im] + element[s_el]*value
    return im


def add_noise(im, u_void=0.8, u_solid=0.2, s_void=0.15, s_solid=0.15):
    r"""
    Add some normally distributed noise values to the image.  This is useful
    for testing binarization routines.

    Parameters
    ----------
    im : ND-array
        The image of the porous media, with 1's or True's denoting voids

    u_solid : float
        The mean greyscale value in the solid space (default is 0.2)

    s_solid : float
        The standard deviation of the greyscale values in the solid phase (the
        default is 0.15)

    u_void: float
        The mean greyscale value in the void space (default is 0.8)

    s_void: float
        The standard deviation of the greyscale values in the void phase (the
        default is 0.15)

    Returns
    -------
    A greyscale image with the same shape as ``im``, but normally distributed
    noise values in the void and solid phase.  A histogram of the default image
    should reveal to distinguishable but overlapping peaks.  This can be
    adjusted using the mean and standard deviation arguments.

    """
    im = im*sp.random.normal(loc=u_void, scale=s_void, size=im.shape) + \
        ~im*sp.random.normal(loc=u_solid, scale=s_solid, size=im.shape)
    return im


def bundle_of_tubes(shape, spacing):
    r"""
    Create a 3D image of a bundle of tubes, in the form of a rectangular
    plate with randomly sized holes through it.

    Parameters
    ----------
    shape : list
        The size the image, with the 3rd dimension indicating the plate
        thickness.  If the 3rd dimension is not given then a thickness of
        1 voxel is assumed.

    spacing : scalar
        The center to center distance of the holes.  The hole sizes will be
        randomly distributed between this values down to 3 voxels.

    Returns
    -------
    A boolean array with True values denoting the pore space
    """
    shape = sp.array(shape)
    if sp.size(shape) == 1:
        shape = sp.full((3, ), int(shape))
    if sp.size(shape) == 2:
        shape = sp.hstack((shape, [1]))
    temp = sp.zeros(shape=shape[:2])
    Xi = sp.ceil(sp.linspace(spacing/2,
                             shape[0]-(spacing/2)-1,
                             shape[0]/spacing))
    Xi = sp.array(Xi, dtype=int)
    Yi = sp.ceil(sp.linspace(spacing/2,
                             shape[1]-(spacing/2)-1,
                             shape[1]/spacing))
    Yi = sp.array(Yi, dtype=int)
    temp[sp.meshgrid(Xi, Yi)] = 1
    inds = sp.where(temp)
    for i in range(len(inds[0])):
        r = sp.random.randint(1, (spacing/2))
        try:
            s1 = slice(inds[0][i]-r, inds[0][i]+r+1)
            s2 = slice(inds[1][i]-r, inds[1][i]+r+1)
            temp[s1, s2] = disk(r)
        except ValueError:
            odd_shape = sp.shape(temp[s1, s2])
            temp[s1, s2] = disk(r)[:odd_shape[0], :odd_shape[1]]
    im = sp.broadcast_to(array=sp.atleast_3d(temp), shape=shape)
    return im


def polydisperse_spheres(shape, porosity, dist, nbins=5):
    r"""
    Create an image of spheres with a distribution of radii.

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny, Nz] where Ni is the
        number of voxels in each direction.  If shape is only 2D, then an
        image of polydisperse disks is returns

    porosity : scalar
        The porosity of the image, defined as the number of void voxels
        divided by the number of voxels in the image. The specified value
        is only matched approximately, so it's suggested to check this value
        after the image is generated.

    dist : scipy.stats distribution object
        This should be an initialized distribution the large number of options
        in the scipy.stats submodule.  For instance, a normal distribution with
        a mean of 20 and a standard deviation of 10 can be obtain with
        ``dist = scipy.stats.norm(loc=20, scale=10)``

    nbins : scalar
        The number of discrete sphere sizes that will be used to generate the
        image.  This function generates  ``nbins`` images of monodisperse
        spheres that span 0.05 and 0.95 of the possible values produced by the
        provided distribution, then overlays them to get polydispersivity.

    Returns
    -------
    A boolean array with True values denoting the pore space
    """
    shape = sp.array(shape)
    if sp.size(shape) == 1:
        shape = sp.full((3, ), int(shape))
    Rs = dist.interval(sp.linspace(0.05, 0.95, nbins))
    Rs = sp.vstack(Rs).T
    Rs = (Rs[:-1] + Rs[1:])/2
    Rs = Rs.flatten()
    phi = 1 - (1 - porosity)/(len(Rs))
    im = sp.ones(shape, dtype=bool)
    for r in Rs:
        temp = overlapping_spheres(shape=shape, radius=r, porosity=phi)
        im = im*temp
    return im


def voronoi_edges(shape, radius, ncells, flat_faces=True):
    r"""
    Create an image of the edges in a Voronoi tessellation

    Parameters
    ----------
    shape : array_like
        The size of the image to generate in [Nx, Ny, Nz] where Ni is the
        number of voxels in each direction.

    radius : scalar
        The radius to which Voronoi edges should be dilated in the final image.

    ncells : scalar
        The number of Voronoi cells to include in the tesselation.

    flat_faces : Boolean
        Whether the Voronoi edges should lie on the boundary of the
        image (True), or if edges outside the image should be removed (False).

    Returns
    -------
    A boolean array with True values denoting the pore space

    """
    shape = sp.array(shape)
    if sp.size(shape) == 1:
        shape = sp.full((3, ), int(shape))
    im = sp.zeros(shape, dtype=bool)
    base_pts = sp.rand(ncells, 3)*shape
    if flat_faces:
        # Reflect base points
        Nx, Ny, Nz = shape
        orig_pts = base_pts
        base_pts = sp.vstack((base_pts, [-1, 1, 1]*orig_pts +
                                        [2.0*Nx, 0, 0]))
        base_pts = sp.vstack((base_pts, [1, -1, 1]*orig_pts +
                                        [0, 2.0*Ny, 0]))
        base_pts = sp.vstack((base_pts, [1, 1, -1]*orig_pts +
                                        [0, 0, 2.0*Nz]))
        base_pts = sp.vstack((base_pts, [-1, 1, 1]*orig_pts))
        base_pts = sp.vstack((base_pts, [1, -1, 1]*orig_pts))
        base_pts = sp.vstack((base_pts, [1, 1, -1]*orig_pts))
    vor = sptl.Voronoi(points=base_pts)
    vor.vertices = sp.around(vor.vertices)
    vor.vertices *= (sp.array(im.shape)-1)/sp.array(im.shape)
    vor.edges = _get_Voronoi_edges(vor)
    for row in vor.edges:
        pts = vor.vertices[row].astype(int)
        if sp.all(pts >= 0) and sp.all(pts < im.shape):
            line_pts = line_segment(pts[0], pts[1])
            im[line_pts] = True
    im = spim.distance_transform_edt(~im) > radius
    return im


def _get_Voronoi_edges(vor):
    r"""
    Given a Voronoi object as produced by the scipy.spatial.Voronoi class,
    this function calculates the start and end points of eeach edge in the
    Voronoi diagram, in terms of the vertex indices used by the received
    Voronoi object.

    Parameters
    ----------
    vor : scipy.spatial.Voronoi object

    Returns
    -------
    A 2-by-N array of vertex indices, indicating the start and end points of
    each vertex in the Voronoi diagram.  These vertex indices can be used to
    index straight into the ``vor.vertices`` array to get spatial positions.
    """
    edges = [[], []]
    for facet in vor.ridge_vertices:
        # Create a closed cycle of vertices that define the facet
        edges[0].extend(facet[:-1]+[facet[-1]])
        edges[1].extend(facet[1:]+[facet[0]])
    edges = sp.vstack(edges).T  # Convert to scipy-friendly format
    mask = sp.any(edges == -1, axis=1)  # Identify edges at infinity
    edges = edges[~mask]  # Remove edges at infinity
    edges = sp.sort(edges, axis=1)  # Move all points to upper triangle
    # Remove duplicate pairs
    edges = edges[:, 0] + 1j*edges[:, 1]  # Convert to imaginary
    edges = sp.unique(edges)  # Remove duplicates
    edges = sp.vstack((sp.real(edges), sp.imag(edges))).T  # Back to real
    edges = sp.array(edges, dtype=int)
    return edges


def circle_pack(shape, radius, offset=0, packing='square'):
    r"""
    Generates a 2D packing of circles

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny] where N is the
        number of voxels in each direction

    radius : scalar
        The radius of circles in the packing (in pixels)

    offset : scalar
        The amount offset (+ or -) to add between pore centers (in pixels).

    packing : string
        Specifies the type of cubic packing to create.  Options are
        'square' (default) and 'triangular'.

    Returns
    -------
    A boolean array with True values denoting the pore space
    """
    r = radius
    shape = sp.array(shape)
    if sp.size(shape) == 1:
        shape = sp.full((2, ), int(shape))
    elif (sp.size(shape) == 3) or (1 in shape):
        raise Exception("This function only produces 2D images, " +
                        "try \'sphere_pack\'")
    im = sp.zeros(shape, dtype=bool)
    if packing.startswith('s'):
        spacing = 2*r
        s = int(spacing/2) + sp.array(offset)
        coords = sp.mgrid[r:im.shape[0]-r:2*s,
                          r:im.shape[1]-r:2*s]
        im[coords[0], coords[1]] = 1
    if packing.startswith('t'):
        spacing = 2*sp.floor(sp.sqrt(2*(r**2))).astype(int)
        s = int(spacing/2) + offset
        coords = sp.mgrid[r:im.shape[0]-r:2*s,
                          r:im.shape[1]-r:2*s]
        im[coords[0], coords[1]] = 1
        coords = sp.mgrid[s+r:im.shape[0]-r:2*s,
                          s+r:im.shape[1]-r:2*s]
        im[coords[0], coords[1]] = 1
    im = spim.distance_transform_edt(~im) >= r
    return im


def sphere_pack(shape, radius, offset=0, packing='sc'):
    r"""
    Generates a cubic packing of spheres

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny, Nz] where N is the
        number of voxels in each direction.

    radius : scalar
        The radius of spheres in the packing

    offset : scalar
        The amount offset (+ or -) to add between pore centers.

    packing : string
        Specifies the type of cubic packing to create.  Options are:

        'sc' : Simple Cubic (default)
        'fcc' : Face Centered Cubic
        'bcc' : Body Centered Cubic

    Returns
    -------
    A boolean array with True values denoting the pore space
    """
    r = radius
    shape = sp.array(shape)
    if sp.size(shape) == 1:
        shape = sp.full((3, ), int(shape))
    elif (sp.size(shape) == 2) or (1 in shape):
        raise Exception("This function only produces 3D images, " +
                        "try \'circle_pack\'")
    im = sp.zeros(shape, dtype=bool)
    if packing.startswith('s'):
        spacing = 2*r
        s = int(spacing/2) + sp.array(offset)
        coords = sp.mgrid[r:im.shape[0]-r:2*s,
                          r:im.shape[1]-r:2*s,
                          r:im.shape[2]-r:2*s]
        im[coords[0], coords[1], coords[2]] = 1
    elif packing.startswith('b'):
        spacing = 2*sp.floor(sp.sqrt(4/3*(r**2))).astype(int)
        s = int(spacing/2) + offset
        coords = sp.mgrid[r:im.shape[0]-r:2*s,
                          r:im.shape[1]-r:2*s,
                          r:im.shape[2]-r:2*s]
        im[coords[0], coords[1], coords[2]] = 1
        coords = sp.mgrid[s+r:im.shape[0]-r:2*s,
                          s+r:im.shape[1]-r:2*s,
                          s+r:im.shape[2]-r:2*s]
        im[coords[0], coords[1], coords[2]] = 1
    elif packing.startswith('f'):
        spacing = 2*sp.floor(sp.sqrt(2*(r**2))).astype(int)
        s = int(spacing/2) + offset
        coords = sp.mgrid[r:im.shape[0]-r:2*s,
                          r:im.shape[1]-r:2*s,
                          r:im.shape[2]-r:2*s]
        im[coords[0], coords[1], coords[2]] = 1
        coords = sp.mgrid[r:im.shape[0]-r:2*s,
                          s+r:im.shape[1]-r:2*s,
                          s+r:im.shape[2]-r:2*s]
        im[coords[0], coords[1], coords[2]] = 1
        coords = sp.mgrid[s+r:im.shape[0]-r:2*s,
                          s:im.shape[1]-r:2*s,
                          s+r:im.shape[2]-r:2*s]
        im[coords[0], coords[1], coords[2]] = 1
        coords = sp.mgrid[s+r:im.shape[0]-r:2*s,
                          s+r:im.shape[1]-r:2*s,
                          s:im.shape[2]-r:2*s]
        im[coords[0], coords[1], coords[2]] = 1
    im = spim.distance_transform_edt(~im) > r
    return im


def overlapping_spheres(shape, radius, porosity):
    r"""
    Generate a packing of overlapping mono-disperse spheres

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny, Nz] where Ni is the
        number of voxels in the *i*th direction.

    radius : scalar
        The radius of spheres in the packing.

    porosity : scalar
        The porosity of the final image.  This number is approximated by
        the method so the returned result may not have exactly the
        specified value.

    Returns
    -------
    A boolean array with True values denoting the pore space

    Notes
    -----
    This method can also be used to generate a dispersion of hollows by
    treating ``porosity`` as solid volume fraction and inverting the
    returned image.
    """
    shape = sp.array(shape)
    if sp.size(shape) == 1:
        shape = sp.full((3, ), int(shape))
    if sp.size(shape) == 2:
        s_vol = sp.sum(disk(radius))
    if sp.size(shape) == 3:
        s_vol = sp.sum(ball(radius))
    bulk_vol = sp.prod(shape)
    N = int(sp.ceil((1 - porosity)*bulk_vol/s_vol))
    im = sp.random.random(size=shape) > (N/bulk_vol)
    im = spim.distance_transform_edt(im) < radius
    return ~im


def noise(shape, porosity=None, octaves=3, frequency=32, mode='simplex'):
    r"""
    Generate a field of spatially correlated random noise using the Perlin
    noise algorithm, or the updated Simplex noise algorithm.

    Parameters
    ----------
    shape : array_like
        The size of the image to generate in [Nx, Ny, Nz] where N is the
        number of voxels.

    porosity : float
        If specified, this will threshold the image to the specified value
        prior to returning.  If no value is given (the default), then the
        scalar noise field is returned.

    octaves : int
        Controls the *texture* of the noise, with higher octaves giving more
        complex features over larger length scales.

    frequency : array_like
        Controls the relative sizes of the features, with higher frequencies
        giving larger features.  A scalar value will apply the same frequency
        in all directions, given an isotropic field; a vector value will
        apply the specified values along each axis to create anisotropy.

    mode : string
        Which noise algorithm to use, either \'simplex\' (default) or
        \'perlin\'.

    Returns
    -------
    If porosity is given, then a boolean array with ``True`` values denoting
    the pore space is returned.  If not, then normally distributed and
    spatially correlated randomly noise is returned.

    Notes
    -----
    This method depends the a package called 'noise' which must be
    compiled. It is included in the Anaconda distribution, or a platform
    specific binary can be downloaded.

    See Also
    --------
    norm_to_uniform

    """
    try:
        import noise
    except:
        raise Exception("The noise package must be installed")
    shape = sp.array(shape)
    if sp.size(shape) == 1:
        Lx, Ly, Lz = sp.full((3, ), int(shape))
    elif len(shape) == 2:
        Lx, Ly = shape
        Lz = 1
    elif len(shape) == 3:
        Lx, Ly, Lz = shape
    if mode == 'simplex':
        f = noise.snoise3
    else:
        f = noise.pnoise3
    frequency = sp.atleast_1d(frequency)
    if frequency.size == 1:
        freq = sp.full(shape=[3, ], fill_value=frequency[0])
    elif frequency.size == 2:
        freq = sp.concatenate((frequency, [1]))
    else:
        freq = sp.array(frequency)
    im = sp.zeros(shape=[Lx, Ly, Lz], dtype=float)
    for x in range(Lx):
        for y in range(Ly):
            for z in range(Lz):
                im[x, y, z] = f(x=x/freq[0], y=y/freq[1], z=z/freq[2],
                                octaves=octaves)
    im = im.squeeze()
    if porosity:
        im = norm_to_uniform(im, scale=[0, 1])
        im = im < porosity
    return im


def blobs(shape, porosity=0.5, blobiness=1):
    """
    Generates an image containing amorphous blobs

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny, Nz] where N is the
        number of voxels

    porosity : float
        If specified, this will threshold the image to the specified value
        prior to returning.  If no value is given (the default), then the
        scalar noise field is returned.

    blobiness : array_like (default = 1)
        Controls the morphology of the blobs.  A higher number results in
        a larger number of small blobs.  If a vector is supplied then the blobs
        are anisotropic.

    Returns
    -------
    If porosity is given, then a boolean array with ``True`` values denoting
    the pore space is returned.  If not, then normally distributed and
    spatially correlated randomly noise is returned.

    See Also
    --------
    norm_to_uniform

    """
    blobiness = sp.array(blobiness)
    shape = sp.array(shape)
    if sp.size(shape) == 1:
        shape = sp.full((3, ), int(shape))
    sigma = sp.mean(shape)/(40*blobiness)
    im = sp.random.random(shape)
    im = spim.gaussian_filter(im, sigma=sigma)
    if porosity:
        im = norm_to_uniform(im, scale=[0, 1])
        im = im < porosity
    return im


def norm_to_uniform(im, scale=None):
    r"""
    Take an image with normally distributed greyscale values and converts it to
    a uniform (i.e. flat) distribution.  It's also possible to specify the
    lower and upper limits of the uniform distribution.

    Parameters
    ----------
    im : ND-image
        The image containing the normally distributed scalar field

    scale : [low, high]
        A list or array indicating the lower and upper bounds for the new
        randomly distributed data.  The default is ``None``, which uses the
        ``max`` and ``min`` of the original image as the the lower and upper
        bounds.

    Returns
    -------
    An ND-image the same size as ``im`` with uniformly distributed greyscale
    values spanning the specified range, if given.
    """
    if scale is None:
        scale = [im.min(), im.max()]
    im = (im - sp.mean(im))/sp.std(im)
    im = 1/2*sp.special.erfc(-im/sp.sqrt(2))
    im = (im - im.min()) / (im.max() - im.min())
    im = im*(scale[1] - scale[0]) + scale[0]
    return im


def cylinders(shape, radius, nfibers, phi_max=0, theta_max=90):
    r"""
    Generates a binary image of overlapping cylinders.  This is a good
    approximation of a fibrous mat.

    Parameters
    ----------
    phi_max : scalar
        A value between 0 and 90 that controls the amount that the fibers
        lie out of the XY plane, with 0 meaning all fibers lie in the XY
        plane, and 90 meaning that fibers are randomly oriented out of the
        plane by as much as +/- 90 degrees.

    theta_max : scalar
        A value between 0 and 90 that controls the amount rotation in the
        XY plane, with 0 meaning all fibers point in the X-direction, and
        90 meaning they are randomly rotated about the Z axis by as much
        as +/- 90 degrees.

    Returns
    -------
    A boolean array with True values denoting the pore space
    """
    shape = sp.array(shape)
    if sp.size(shape) == 1:
        shape = sp.full((3, ), int(shape))
    elif sp.size(shape) == 2:
        raise Exception("2D fibers don't make sense")
    im = sp.zeros(shape)
    R = sp.sqrt(sp.sum(sp.square(shape)))
    n = 0
    while n < nfibers:
        x = sp.rand(3)*shape
        phi = sp.deg2rad(90 + 90*(0.5 - sp.rand())*phi_max/90)
        theta = sp.deg2rad(180 - 90*(0.5 - sp.rand())*2*theta_max/90)
        X0 = R*sp.array([sp.sin(theta)*sp.cos(phi),
                         sp.sin(theta)*sp.sin(phi),
                         sp.cos(theta)])
        [X0, X1] = [X0 + x, -X0 + x]
        crds = line_segment(X0, X1)
        lower = ~sp.any(sp.vstack(crds).T < [0, 0, 0], axis=1)
        upper = ~sp.any(sp.vstack(crds).T >= shape, axis=1)
        valid = upper*lower
        if sp.any(valid):
            im[crds[0][valid], crds[1][valid], crds[2][valid]] = 1
            n += 1
    im = sp.array(im, dtype=bool)
    dt = spim.distance_transform_edt(~im) < radius
    return ~dt


def line_segment(X0, X1):
    r"""
    Calculate the voxel coordinates of a straight line between the two given
    end points

    Parameters
    ----------
    X0 and X1 : array_like
        The [x, y, z] coordinates of the start and end points of the line.

    Returns
    -------
        A list of lists containing the X, Y, and Z coordinates of all voxels
        that should be drawn between the start and end points to create a solid
        line.
    """
    X0 = sp.around(X0)
    X1 = sp.around(X1)
    L = sp.amax(sp.absolute([[X1[0]-X0[0]], [X1[1]-X0[1]], [X1[2]-X0[2]]])) + 1
    x = sp.rint(sp.linspace(X0[0], X1[0], L)).astype(int)
    y = sp.rint(sp.linspace(X0[1], X1[1], L)).astype(int)
    z = sp.rint(sp.linspace(X0[2], X1[2], L)).astype(int)
    return [x, y, z]



r'''
=====
Tools
=====

This module contains a variety of functions for manipulating images in ways
that do NOT return a modified version of the original image.

.. autofunction:: extend_slice
.. autofunction:: extract_subsection
.. autofunction:: extract_cylinder
.. autofunction:: find_outer_region
.. autofunction:: get_border
.. autofunction:: get_slice
.. autofunction:: make_contiguous
.. autofunction:: randomize_colors
.. autofunction:: subdivide

'''

from .__funcs__ import binary_opening_fast
from .__funcs__ import extend_slice
from .__funcs__ import extract_subsection
from .__funcs__ import extract_cylinder
from .__funcs__ import find_outer_region
from .__funcs__ import get_border
from .__funcs__ import get_slice
from .__funcs__ import make_contiguous
from .__funcs__ import randomize_colors
from .__funcs__ import subdivide



r"""
This module contains functions for extracting pore networks from images.

.. autofunction:: partition_pore_space
.. autofunction:: align_image_with_openpnm
.. autofunction:: snow
.. autofunction:: trim_saddle_points
.. autofunction:: trim_nearby_peaks
.. autofunction:: reduce_peaks_to_points
.. autofunction:: find_peaks
.. autofunction:: extract_pore_network

"""

from .__funcs__ import partition_pore_space
from .__funcs__ import align_image_with_openpnm
from .__snow__ import snow
from .__snow__ import trim_saddle_points
from .__snow__ import trim_nearby_peaks
from .__snow__ import reduce_peaks_to_points
from .__snow__ import find_peaks
from .__getnet__ import extract_pore_network



import numpy as np
from scipy import ndimage as spim
from porespy.io.evtk import hl as bp
import scipy.ndimage as nd


def to_vtk(im, path='./voxvtk', divide=False, downsample=False, voxel_size=1,
           vox=False):
    r"""
    Wrapper for the pyevtk
    Copyright 2010 - 2016 Paulo A. Herrera. All rights reserved. (see /evtk
    folder for complete license information)

    Parameters
    ----------
    im : 3D image
        The image of the porous material

    path : string
        Path to output file

    divide : bool
        vtk files can get very large, this option allows you for two output
        files, divided at z = half. This allows for large data sets to be
        imaged without loss of information

    downsample : bool
        very large images acan be downsampled to half the size in each
        dimension, this doubles the effective voxel size

    voxel_size : int
        The side length of the voxels (voxels  are cubic)

    vox : bool
        For an image that is binary (1's and 0's) this reduces the file size by
        using int8 format (can also be used to reduce file size when accuracy
        is not necessary ie: just visulization)

    Output
    ------
    File: vtk, vtp or vti file that can opened in paraview
    """
    if vox:
        im = im.astype(np.int8)
    vs = voxel_size
    if divide:
        split = np.round(im.shape[2]/2).astype(np.int)
        im1 = im[:, :, 0:split]
        im2 = im[:, :, split:]
        bp.imageToVTK(path+'1', cellData={'im': np.ascontiguousarray(im1)},
                      spacing=(vs, vs, vs))
        bp.imageToVTK(path+'2', origin=(0.0, 0.0, split*vs),
                      cellData={'im': np.ascontiguousarray(im2)},
                      spacing=(vs, vs, vs))
    elif downsample:
        im = spim.interpolation.zoom(im, zoom=0.5, order=0, mode='reflect')
        bp.imageToVTK(path, cellData={'im': np.ascontiguousarray(im)},
                      spacing=(2*vs, 2*vs, 2*vs))
    else:
        bp.imageToVTK(path, cellData={'im': np.ascontiguousarray(im)},
                      spacing=(vs, vs, vs))


def to_palabos(im, filename, solid=0):
    r"""
    Converts an ND-array image to a text file that Palabos can read in as a
    geometry for Lattice Boltzmann simulations. Uses a Euclidean distance
    transform to identify solid voxels neighboring fluid voxels and labels
    them as the interface.

    Parameters
    ----------
    im : ND-array
        The image of the porous material

    filename : string
        Path to output file

    solid : int
        The value of the solid voxels in the image used to convert image to
        binary with all other voxels assumed to be fluid.

    Output
    -------
    File produced contains 3 values: 2 = Solid, 1 = Interface, 0 = Pore

    """
    # Create binary image for fluid and solid phases
    bin_im = im == solid
    # Transform to integer for distance transform
    bin_im = bin_im.astype(int)
    # Distance Transform computes Euclidean distance in lattice units to
    # Nearest fluid for every solid voxel
    dt = nd.distance_transform_edt(bin_im)
    dt[dt > np.sqrt(2)] = 2
    dt[(dt > 0)*(dt <= np.sqrt(2))] = 1
    dt = dt.astype(int)
    # Write out data
    with open(filename, 'w') as f:
        out_data = dt.flatten().tolist()
        f.write('\n'.join(map(repr, out_data)))



import scipy as sp


def sem(im, direction='X'):
    r"""
    Simulates an SEM photograph looking into the porous material in the
    specified direction.  Features are colored according to their depth into
    the image, so darker features are further away.

    Parameters
    ----------
    im : array_like
        ND-image of the porous material with the solid phase marked as 1 or
        True

    direction : string
        Specify the axis along which the camera will point.  Options are
        'X', 'Y', and 'Z'.

    Returns
    -------
    A 2D greyscale image suitable for use in matplotlib\'s ```imshow```
    function.
    """
    im = sp.array(~im, dtype=int)
    if direction in ['Y', 'y']:
        im = sp.transpose(im, axes=[1, 0, 2])
    if direction in ['Z', 'z']:
        im = sp.transpose(im, axes=[2, 1, 0])
    t = im.shape[0]
    depth = sp.reshape(sp.arange(0, t), [t, 1, 1])
    im = im*depth
    im = sp.amax(im, axis=0)
    return im


def xray(im, direction='X'):
    r"""
    Simulates an X-ray radiograph looking through the porouls material in the
    specfied direction.  The resulting image is colored according to the amount
    of attenuation an X-ray would experience, so regions with more solid will
    appear darker.

    Parameters
    ----------
    im : array_like
        ND-image of the porous material with the solid phase marked as 1 or
        True

    direction : string
        Specify the axis along which the camera will point.  Options are
        'X', 'Y', and 'Z'.

    Returns
    -------
    A 2D greyscale image suitable for use in matplotlib\'s ```imshow```
    function.
    """
    im = sp.array(~im, dtype=int)
    if direction in ['Y', 'y']:
        im = sp.transpose(im, axes=[1, 0, 2])
    if direction in ['Z', 'z']:
        im = sp.transpose(im, axes=[2, 1, 0])
    im = sp.sum(im, axis=0)
    return im



import scipy as sp
import numpy as np
from skimage.segmentation import clear_border
from skimage.feature import peak_local_max
import scipy.ndimage as spim
import scipy.spatial as sptl
from porespy.tools import get_border, extract_subsection, extend_slice
from collections import namedtuple
from tqdm import tqdm
from scipy import fftpack as sp_ft


def representative_elementary_volume(im, npoints=1000):
    r"""
    Calculates the porosity of the image as a function subdomain size.  This
    function extracts a specified number of subdomains of random size, then
    finds their porosity.

    Parameters
    ----------
    im : ND-array
        The image of the porous material

    npoints : int
        The number of randomly located and sized boxes to sample.  The default
        is 1000.

    Returns
    -------
    A tuple containing the ND-arrays: The subdomain *volume* and its
    *porosity*.  Each of these arrays is ``npoints`` long.  They can be
    conveniently plotted by passing the tuple to matplotlib's ``plot`` function
    using the \* notation: ``plt.plot(*the_tuple, 'b.')``.  The resulting plot
    is similar to the sketch given by Bachmat and Bear [1]

    Notes
    -----
    This function is frustratingly slow.  Profiling indicates that all the time
    is spent on scipy's ``sum`` function which is needed to sum the number of
    void voxels (1's) in each subdomain.

    Also, this function is primed for parallelization since the ``npoints`` are
    calculated independenlty.

    References
    ----------
    [1] Bachmat and Bear. On the Concept and Size of a Representative
    Elementary Volume (Rev), Advances in Transport Phenomena in Porous Media
    (1987)

    """
    im_temp = sp.zeros_like(im)
    crds = sp.array(sp.rand(npoints, im.ndim)*im.shape, dtype=int)
    pads = sp.array(sp.rand(npoints)*sp.amin(im.shape)/2+10, dtype=int)
    im_temp[tuple(crds.T)] = True
    labels, N = spim.label(input=im_temp)
    slices = spim.find_objects(input=labels)
    porosity = sp.zeros(shape=(N,), dtype=float)
    volume = sp.zeros(shape=(N,), dtype=int)
    for i in tqdm(sp.arange(0, N)):
        s = slices[i]
        p = pads[i]
        new_s = extend_slice(s, shape=im.shape, pad=p)
        temp = im[new_s]
        Vp = sp.sum(temp)
        Vt = sp.size(temp)
        porosity[i] = Vp/Vt
        volume[i] = Vt
    profile = namedtuple('profile', ('volume', 'porosity'))
    profile.volume = volume
    profile.porosity = porosity
    return profile


def radial_distribution(im, bins=10):
    r"""

    References
    ----------
    [1] Torquato, S. Random Heterogeneous Materials: Mircostructure and
    Macroscopic Properties. Springer, New York (2002)
    """
    print("Sorry, but this function is not implemented yet")


def lineal_path(im):
    r"""

    References
    ----------
    [1] Torquato, S. Random Heterogeneous Materials: Mircostructure and
    Macroscopic Properties. Springer, New York (2002)
    """
    print("Sorry, but this function is not implemented yet")


def pore_size_density(im, bins=10, voxel_size=1):
    r"""
    Computes the histogram of the distance transform as an estimator of the
    pore sizes in the image.

    Parameters
    ----------
    im : ND-array
        Either a binary image of the pore space with ``True`` indicating the
        pore phase (or phase of interest), or a pre-calculated distance
        transform which can save time.

    bins : int or array_like
        This number of bins (if int) or the location of the bins (if array).
        This argument is passed directly to Scipy's ``histogram`` function so
        see that docstring for more information.  The default is 10 bins, which
        reduces produces a relatively smooth distribution.

    voxel_size : scalar
        The size of a voxel side in preferred units.  The default is 1, so the
        user can apply the scaling to the returned results after the fact.

    Returns
    -------
    A tuple containing two 1D arrays: ``radius `` is the radius of the voxels,
    and ``count`` is the number of voxels that are within R of the solid.

    Notes
    -----
    This function should not be taken as a pore size distribution in the
    explict sense, but rather an indicator of the sizes in the image.

    References
    ----------
    [1] Torquato, S. Random Heterogeneous Materials: Mircostructure and
    Macroscopic Properties. Springer, New York (2002) - See page 292
    """
    if im.dtype == bool:
        im = spim.distance_transform_edt(im)
    hist = sp.histogram(a=im[im > 0], bins=bins)
    n = hist[0]/sp.sum(im > 0)
    r = hist[1][:-1]*voxel_size
    rdf = namedtuple('rdf', ('radius', 'count'))
    return rdf(r, n)


def porosity(im):
    r"""
    Calculates the porosity of an image assuming 1's are void space and 0's are
    solid phase.  All other values are ignored.

    Parameters
    ----------
    im : ND-array
        Image of the void space with 1's indicating void space (or True) and
        0's indicating the solid phase (or False).

    Returns
    -------
    porosity : float
        Calculated as the sum of all 1's divided by the sum of all 1's and 0's.

    Notes
    -----
    This function assumes void is represented by 1 and solid by 0, and all
    other values are ignored.  This is useful, for example, for images of
    cylindrical cores, where all voxels outside the core are labelled with 2.

    Alternatively, images can be processed with ``find_disconnected_voxels``
    to get an image of only blind pores.  This can then be added to the orignal
    image such that blind pores have a value of 2, thus allowing the
    calculation of accessible porosity, rather than overall porosity.

    """
    im = sp.array(im, dtype=int)
    Vp = sp.sum(im == 1)
    Vs = sp.sum(im == 0)
    e = Vp/(Vs + Vp)
    return e


def two_point_correlation_bf(im, spacing=10):
    r"""
    Calculates the two-point correlation function using brute-force (see Notes)

    Parameters
    ----------
    im : ND-array
        The image of the void space on which the 2-point correlation is desired

    spacing : int
        The space between points on the regular grid that is used to generate
        the correlation (see Notes)

    Returns
    -------
    A tuple containing the x and y data for plotting the two-point correlation
    function, using the *args feature of matplotlib's plot function.  The x
    array is the distances between points and the y array is corresponding
    probabilities that points of a given distance both lie in the void space.

    The distance values are binned as follows:

        bins = range(start=0, stop=sp.amin(im.shape)/2, stride=spacing)

    Notes
    -----
    The brute-force approach means overlaying a grid of equally spaced points
    onto the image, calculating the distance between each and every pair of
    points, then counting the instances where both pairs lie in the void space.

    This approach uses a distance matrix so can consume memory very quickly for
    large 3D images and/or close spacing.
    """
    if im.ndim == 2:
        pts = sp.meshgrid(range(0, im.shape[0], spacing),
                          range(0, im.shape[1], spacing))
        crds = sp.vstack([pts[0].flatten(),
                          pts[1].flatten()]).T
    elif im.ndim == 3:
        pts = sp.meshgrid(range(0, im.shape[0], spacing),
                          range(0, im.shape[1], spacing),
                          range(0, im.shape[2], spacing))
        crds = sp.vstack([pts[0].flatten(),
                          pts[1].flatten(),
                          pts[2].flatten()]).T
    dmat = sptl.distance.cdist(XA=crds, XB=crds)
    hits = im[pts].flatten()
    dmat = dmat[hits, :]
    h1 = sp.histogram(dmat, bins=range(0, int(sp.amin(im.shape)/2), spacing))
    dmat = dmat[:, hits]
    h2 = sp.histogram(dmat, bins=h1[1])
    tpcf = namedtuple('two_point_correlation_function',
                      ('distance', 'probability'))
    return tpcf(h2[1][:-1], h2[0]/h1[0])


def _radial_profile(autocorr, r_max, nbins=100):
    r"""
    Helper functions to calculate the radial profile of the autocorrelation
    Masks the image in radial segments from the center and averages the values
    The distance values are normalized and 100 bins are used as default.

    Parameters
    ----------
    autocorr : ND-array
        The image of autocorrelation produced by FFT

    r_max : int or float
        The maximum radius in pixels to sum the image over
    """
    if len(autocorr.shape) == 2:
        adj = sp.reshape(autocorr.shape, [2, 1, 1])
        inds = sp.indices(autocorr.shape) - adj/2
        dt = sp.sqrt(inds[0]**2 + inds[1]**2)
    elif len(autocorr.shape) == 3:
        adj = sp.reshape(autocorr.shape, [3, 1, 1, 1])
        inds = sp.indices(autocorr.shape) - adj/2
        dt = sp.sqrt(inds[0]**2 + inds[1]**2 + inds[2]**2)
    else:
        raise Exception('Image dimensions must be 2 or 3')
    bin_size = np.int(np.ceil(r_max/nbins))
    bins = np.arange(bin_size, r_max, step=bin_size)
    radial_sum = np.zeros_like(bins)
    for i, r in enumerate(bins):
        # Generate Radial Mask from dt using bins
        mask = (dt <= r) * (dt > (r-bin_size))
        radial_sum[i] = np.sum(autocorr[mask])/np.sum(mask)
    # Return normalized bin and radially summed autoc
    norm_autoc_radial = radial_sum/np.max(autocorr)
    tpcf = namedtuple('two_point_correlation_function',
                      ('distance', 'probability'))
    return tpcf(bins, norm_autoc_radial)


def two_point_correlation_fft(im):
    r"""
    Calculates the two-point correlation function using fourier transforms

    Parameters
    ----------
    im : ND-array
        The image of the void space on which the 2-point correlation is desired

    Returns
    -------
    A tuple containing the x and y data for plotting the two-point correlation
    function, using the *args feature of matplotlib's plot function.  The x
    array is the distances between points and the y array is corresponding
    probabilities that points of a given distance both lie in the void space.

    Notes
    -----
    The fourier transform approach utilizes the fact that the autocorrelation
    function is the inverse FT of the power spectrum density.
    For background read the Scipy fftpack docs and for a good explanation see:
    http://www.ucl.ac.uk/~ucapikr/projects/KamilaSuankulova_BSc_Project.pdf
    """
    # Calculate half lengths of the image
    hls = (np.ceil(np.shape(im))/2).astype(int)
    # Fourier Transform and shift image
    F = sp_ft.ifftshift(sp_ft.fftn(sp_ft.fftshift(im)))
    # Compute Power Spectrum
    P = sp.absolute(F**2)
    # Auto-correlation is inverse of Power Spectrum
    autoc = sp.absolute(sp_ft.ifftshift(sp_ft.ifftn(sp_ft.fftshift(P))))
    tpcf = _radial_profile(autoc, r_max=np.min(hls))
    return tpcf


def pore_size_distribution(im):
    r"""
    Calculate drainage curve based on the image produced by the
    ``porosimetry`` function.

    Parameters
    ----------
    im : ND-array
        The array of entry sizes as produced by the ``porosimetry`` function.

    Returns
    -------
    Rp, Snwp: Two arrays containing (a) the radius of the penetrating
    sphere (in voxels) and (b) the volume fraction of pore phase voxels
    that are accessible from the specfied inlets.

    Notes
    -----
    This function normalizes the invading phase saturation by total pore
    volume of the dry image, which is assumed to be all voxels with a value
    equal to 1.  To do porosimetry on images with large outer regions,
    use the ```find_outer_region``` function then set these regions to 0 in
    the input image.  In future, this function could be adapted to apply
    this check by default.

    """
    sizes = sp.unique(im)
    R = []
    Snwp = []
    Vp = sp.sum(im > 0)
    for r in sizes[1:]:
        R.append(r)
        Snwp.append(sp.sum(im >= r))
    Snwp = [s/Vp for s in Snwp]
    data = namedtuple('xy_data', ('radius', 'saturation'))
    return data(R, Snwp)


def chord_length_counts(im):
    r"""
    Determines the length of each chord in the supplied image by looking at
    its size.

    Parameters
    ----------
    im : ND-array
        An image containing chords drawn in the void space.

    Returns
    -------
    A 1D array with one element for each chord, containing its length.

    Notes
    ----
    The returned array can be passed to ``plt.hist`` to plot the histogram,
    or to ``sp.histogram`` to get the histogram data directly. Another useful
    function is ``sp.bincount`` which gives the number of chords of each
    length in a format suitable for ``plt.plot``.
    """
    labels, N = spim.label(im > 0)
    slices = spim.find_objects(labels)
    chord_lens = sp.zeros(N, dtype=int)
    for i in range(len(slices)):
        s = slices[i]
        chord_lens[i] = sp.amax([item.stop-item.start for item in s])
    return chord_lens


def chord_length_distribution(im, bins=25, log=False):
    r"""
    Determines the distribution of chord lengths in a image containing chords.

    Parameters
    ----------
    im : ND-image
        An image with chords drawn in the pore space, as produced by
        ``apply_chords`` or ``apply_chords_3d``.

    bins : scalar or array_like
        If a scalar is given it is interpreted as the number of bins to use,
        and if an array is given they are used as the bins directly.

    log : Boolean
        If true, the logarithm of the chord lengths will be used, which can
        make the data more clear.

    Returns
    -------
    A tuple containing the ``chord_length_bins``, and four separate pieces of
    information: ``cumulative_chord_count`` and ``cumulative_chord_length``,
    as well as the ``differenial_chord_count`` and
    ``differential_chord_length``.
    """
    h = chord_length_counts(im)
    if log:
        h = sp.log10(h)
    y_num, x = sp.histogram(h, bins=bins, density=True)
    y_len, x = sp.histogram(h, bins=bins, weights=h, density=True)
    y_num_cum = sp.cumsum((y_num*(x[1:]-x[:-1]))[::-1])[::-1]
    y_len_cum = sp.cumsum((y_len*(x[1:]-x[:-1]))[::-1])[::-1]
    data = namedtuple('chord_distribution', ('chord_length_bins',
                                             'cumulative_chord_count',
                                             'cumulative_chord_length',
                                             'differential_chord_count',
                                             'differential_chord_length'))
    return data(x[:-1], y_num_cum, y_len_cum, y_num, y_len)



# ***********************************************************************************
# Copyright 2010 - 2016 Paulo A. Herrera. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2.Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY PAULO A. HERRERA ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING
# BUT NOT LIMITED TO, PROCUREMEN OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ***********************************************************************************

# **************************************
# *  Simple class to generate a        *
# *  well-formed XML file.             *
# **************************************


class XmlWriter:
    def __init__(self, filepath, addDeclaration=True):
        self.stream = open(filepath, "wb")
        self.openTag = False
        self.current = []
        if (addDeclaration):
            self.addDeclaration()

    def close(self):
        assert(not self.openTag)
        self.stream.close()

    def addDeclaration(self):
        self.stream.write(b'<?xml version="1.0"?>')

    def openElement(self, tag):
        if self.openTag:
            self.stream.write(b">")
        st = "\n<%s" % tag
        self.stream.write(str.encode(st))
        self.openTag = True
        self.current.append(tag)
        return self

    def closeElement(self, tag=None):
        if tag:
            assert(self.current.pop() == tag)
            if (self.openTag):
                self.stream.write(b">")
                self.openTag = False
            st = "\n</%s>" % tag
            self.stream.write(str.encode(st))
        else:
            self.stream.write(b"/>")
            self.openTag = False
            self.current.pop()
        return self

    def addText(self, text):
        if (self.openTag):
            self.stream.write(b">\n")
            self.openTag = False
        self.stream.write(str.encode(text))
        return self

    def addAttributes(self, **kwargs):
        assert (self.openTag)
        for key in kwargs:
            st = ' %s="%s"' % (key, kwargs[key])
            self.stream.write(str.encode(st))
        return self



import scipy as sp
from skimage.segmentation import clear_border
import scipy.ndimage as spim
import scipy.spatial as sptl
from porespy.tools import get_border, extract_subsection, extend_slice
from collections import namedtuple
from tqdm import tqdm
from numba import jit
from skimage.morphology import ball, disk, square, cube
from skimage.morphology import reconstruction


def find_disconnected_voxels(im, conn=None):
    r"""
    This identifies all pore (or solid) voxels that are not connected to the
    edge of the image.  This can be used to find blind pores, or remove
    artifacts such as solid phase voxels that are floating in space.

    Parameters
    ----------
    im : ND-image
        A Boolean image, with True values indicating the phase for which
        disconnected voxels are sought.

    conn : int
        For 2D the options are 4 and 8 for square and diagonal neighbors, while
        for the 3D the options are 6 and 26, similarily for square and diagonal
        neighbors.  The default is max

    Returns
    -------
    An ND-image the same size as ``im``, with True values indicating voxels of
    the phase of interest (i.e. True values in the original image) that are
    not connected to the outer edges.

    Notes
    -----
    The returned array (e.g. ``holes``) be used to trim blind pores from
    ``im`` using: ``im[holes] = False``

    """
    if im.ndim == 2:
        if conn == 4:
            strel = disk(1)
        elif conn in [None, 8]:
            strel = square(3)
    elif im.ndim == 3:
        if conn == 6:
            strel = ball(1)
        elif conn in [None, 26]:
            strel = cube(3)
    labels, N = spim.label(input=im, structure=strel)
    holes = clear_border(labels=labels) > 0
    return holes


def fill_blind_pores(im):
    r"""
    Fills all pores that are not connected to the edges of the image.

    Parameters
    ----------
    im : ND-array
        The image of the porous material

    Returns
    -------
        A version of ``im`` but with all the disconnected pores removed.

    See Also
    --------
    find_disconnected_voxels

    """
    holes = find_disconnected_voxels(im)
    im[holes] = False
    return im


def trim_floating_solid(im):
    r"""
    Removes all solid that that is not attached to the edges of the image.

    Parameters
    ----------
    im : ND-array
        The image of the porous material

    Returns
    -------
        A version of ``im`` but with all the disconnected solid removed.

    See Also
    --------
    find_disconnected_voxels

    """
    holes = find_disconnected_voxels(~im)
    im[holes] = True
    return im


def trim_extrema(im, h, mode='maxima'):
    r"""
    This trims local extrema in greyscale values by a specified amount,
    essentially decapitating peaks or flooding valleys, or both.

    Parameters
    ----------
    im : ND-array
        The image whose extrema are to be removed

    h : scalar
        The height to remove from each peak or fill in each valley

    mode : string {'maxima' | 'minima' | 'extrema'}
        Specifies whether to remove maxima or minima or both

    Returns
    -------
    A copy of the input image with all the peaks and/or valleys removed.

    Notes
    -----
    This function is referred to as **imhmax** or **imhmin** in Matlab.
    """
    result = im
    if mode in ['maxima', 'extrema']:
        result = reconstruction(seed=im - h, mask=im, method='dilation')
    elif mode in ['minima', 'extrema']:
        result = reconstruction(seed=im + h, mask=im, method='erosion')
    return result


@jit
def flood(im, mode='max'):
    r"""
    Floods/fills each region in an image with a single value based on the
    specific values in that region.  The ``mode`` argument is used to
    determine how the value is calculated.  A region is defined as a connected
    cluster of voxels surrounded by 0's for False's.

    Parameters
    ----------
    im : array_like
        An ND image with isolated regions containing 0's elsewhere.

    mode : string
        Specifies how to determine which value should be used to flood each
        region.  Options are:

    *'max'* : Floods each region with the local maximum in that region

    *'min'* : Floods each region the local minimum in that region

    *'size'* : Floods each region with the size of that region

    Returns
    -------
    An ND-array the same size as ``im`` with new values placed in each
    forground voxel based on the ``mode``.

    """
    labels, N = spim.label(im)
    mask = im != 0
    I = im.flatten()
    L = labels.flatten()
    if mode.startswith('max'):
        V = sp.zeros(shape=N+1, dtype=float)
        for i in range(len(L)):
            if V[L[i]] < I[i]:
                V[L[i]] = I[i]
    elif mode.startswith('min'):
        V = sp.ones(shape=N+1, dtype=float)*sp.inf
        for i in range(len(L)):
            if V[L[i]] > I[i]:
                V[L[i]] = I[i]
    elif mode.startswith('size'):
        V = sp.zeros(shape=N+1, dtype=int)
        for i in range(len(L)):
            V[L[i]] += 1
    im_flooded = sp.reshape(V[labels], newshape=im.shape)
    im_flooded = im_flooded*mask
    return im_flooded


def apply_chords(im, spacing=0, axis=0, trim_edges=True):
    r"""
    Adds chords to the void space in the specified direction.  The chords are
    separated by 1 voxel plus the provided spacing.

    Parameters
    ----------
    im : ND-array
        An image of the porous material with void marked as True.

    spacing : int (default = 0)
        Chords are automatically separated by 1 voxel and this argument
        increases the separation.

    axis : int (default = 0)
        The axis along which the chords are drawn.

    trim_edges : bool (default = True)
        Whether or not to remove chords that touch the edges of the image.
        These chords are artifically shortened, so skew the chord length
        distribution

    Returns
    -------
    An ND-array of the same size as ```im``` with True values indicating the
    chords.

    See Also
    --------
    apply_chords_3D

    """
    if spacing < 0:
        raise Exception('Spacing cannot be less than 0')
    dims1 = sp.arange(0, im.ndim)
    dims2 = sp.copy(dims1)
    dims2[axis] = 0
    dims2[0] = axis
    im = sp.moveaxis(a=im, source=dims1, destination=dims2)
    im = sp.atleast_3d(im)
    ch = sp.zeros_like(im, dtype=bool)
    if im.ndim == 2:
        ch[:, ::2+spacing, ::2+spacing] = 1
    if im.ndim == 3:
        ch[:, ::4+2*spacing, ::4+2*spacing] = 1
    chords = im*ch
    chords = sp.squeeze(chords)
    if trim_edges:
        temp = clear_border(spim.label(chords == 1)[0]) > 0
        chords = temp*chords
    chords = sp.moveaxis(a=chords, source=dims1, destination=dims2)
    return chords


def apply_chords_3D(im, spacing=0, trim_edges=True):
    r"""
    Adds chords to the void space in all three principle directions.  The
    chords are seprated by 1 voxel plus the provided spacing.  Chords in the X,
    Y and Z directions are labelled 1, 2 and 3 resepctively.

    Parameters
    ----------
    im : ND-array
        A 3D image of the porous material with void space marked as True.

    spacing : int (default = 0)
        Chords are automatically separed by 1 voxel on all sides, and this
        argument increases the separation.

    trim_edges : bool (default = True)
        Whether or not to remove chords that touch the edges of the image.
        These chords are artifically shortened, so skew the chord length
        distribution

    Returns
    -------
    An ND-array of the same size as ```im``` with values of 1 indicating
    x-direction chords, 2 indicating y-direction chords, and 3 indicating
    z-direction chords.

    Notes
    -----
    The chords are separated by a spacing of at least 1 voxel so that tools
    that search for connected components, such as ``scipy.ndimage.label`` can
    detect individual chords.

    See Also
    --------
    apply_chords

    """
    if im.ndim < 3:
        raise Exception('Must be a 3D image to use this function')
    if spacing < 0:
        raise Exception('Spacing cannot be less than 0')
    ch = sp.zeros_like(im, dtype=int)
    ch[:, ::4+2*spacing, ::4+2*spacing] = 1  # X-direction
    ch[::4+2*spacing, :, 2::4+2*spacing] = 2  # Y-direction
    ch[2::4+2*spacing, 2::4+2*spacing, :] = 3  # Z-direction
    chords = ch*im
    if trim_edges:
        temp = clear_border(spim.label(chords > 0)[0]) > 0
        chords = temp*chords
    return chords


def local_thickness(im):
    r"""
    For each voxel, this functions calculates the radius of the largest sphere
    that both engulfs the voxel and fits entirely within the foreground. This
    is not the same as a simple distance transform, which finds the largest
    sphere that could be *centered* on each voxel.

    Parameters
    ----------
    im : array_like
        A binary image with the phase of interest set to True

    Returns
    -------
    An image with the pore size values in each voxel

    Notes
    -----
    The term *foreground* is used since this function can be applied to both
    pore space or the solid, whichever is set to True.

    """
    from skimage.morphology import cube
    if im.ndim == 2:
        from skimage.morphology import square as cube
    dt = spim.distance_transform_edt(im)
    sizes = sp.unique(sp.around(dt, decimals=0))
    im_new = sp.zeros_like(im, dtype=float)
    for r in tqdm(sizes):
        im_temp = dt >= r
        im_temp = spim.distance_transform_edt(~im_temp) <= r
        im_new[im_temp] = r
    # Trim outer edge of features to remove noise
    im_new = spim.binary_erosion(input=im, structure=cube(1))*im_new
    return im_new


def porosimetry(im, npts=25, sizes=None, inlets=None, access_limited=True):
    r"""
    Performs a porosimetry simulution on the image

    Parameters
    ----------
    im : ND-array
        An ND image of the porous material containing True values in the
        pore space.

    npts : scalar
        The number of invasion points to simulate.  Points will be
        generated spanning the range of sizes in the distance transform.
        The default is 25 points.

    sizes : array_like
        The sizes to invade.  Use this argument instead of ``npts`` for
        more control of the range and spacing of points.

    inlets : ND-array, boolean
        A boolean mask with True values indicating where the invasion
        enters the image.  By default all faces are considered inlets,
        akin to a mercury porosimetry experiment.  Users can also apply
        solid boundaries to their image externally before passing it in,
        allowing for complex inlets like circular openings, etc.

    access_limited : Boolean
        This flag indicates if the intrusion should only occur from the
        surfaces (``access_limited`` is True, which is the default), or
        if the invading phase should be allowed to appear in the core of
        the image.  The former simulates experimental tools like mercury
        intrusion porosimetry, while the latter is useful for comparison
        to gauge the extent of shielding effects in the sample.

    Returns
    -------
    An ND-image with voxel values indicating the sphere radius at which it
    becomes accessible from the inlets.  This image can be used to find
    invading fluid configurations as a function of applied capillary pressure
    by applying a boolean comparison: ``inv_phase = im > r`` where ``r`` is
    the radius (in voxels) of the invading sphere.  Of course, ``r`` can be
    converted to capillary pressure using your favorite model.

    """
    dt = spim.distance_transform_edt(im > 0)
    if inlets is None:
        inlets = get_border(im.shape, mode='faces')
    inlets = sp.where(inlets)
    if sizes is None:
        sizes = sp.logspace(start=sp.log10(sp.amax(dt)), stop=0, num=npts)
    else:
        sizes = sp.sort(a=sizes)[-1::-1]
    imresults = sp.zeros(sp.shape(im))
    for r in tqdm(sizes):
        imtemp = dt >= r
        if access_limited:
            imtemp[inlets] = True  # Add inlets before labeling
            labels, N = spim.label(imtemp)
            imtemp = imtemp ^ (clear_border(labels=labels) > 0)
            imtemp[inlets] = False  # Remove inlets
        if sp.any(imtemp):
            imtemp = spim.distance_transform_edt(~imtemp) < r
            imresults[(imresults == 0)*imtemp] = r
    return imresults



from collections import namedtuple
import scipy as sp
import scipy.ndimage as spim
import scipy.spatial as sptl
from skimage.morphology import disk, ball, square, cube
from porespy.tools import extend_slice, randomize_colors
from porespy.network_extraction import partition_pore_space


def snow(im, r_max=4, sigma=0.4):
    r"""
    This function partitions the void space into pore regions using a
    marker-based watershed algorithm.  The key to this function is that true
    local maximum of the distance transform are found by trimming various
    types of extraneous peaks.

    The SNOW network extraction algorithm (Sub-Network of an Over-segmented
    Watershed) was designed to handle to perculiarities of high porosity
    materials, but it applies well to other materials as well.

    Parameters
    ----------
    im : array_like
        Can be either (a) a boolean image of the domain, with ``True``
        indicating the pore space and ``False`` elsewhere, or (b) a distance
        transform of the domain calculated externally by the user.  Option (b)
        is faster if a distance transform is already available.

    r_max : scalar
        The radius of there spherical structuring element to use in the Maximum
        filter stage that is used to find peaks.  The default is 4

    sigma : scalar
        The standard deviation of the Gaussian filter used in step 1.  The
        default is 0.4.  If 0 is given then the filter is not applied, which is
        useful if a distance transform is supplied as the ``im`` argument that
        has already been processed.

    Returns
    -------
    A tuple containing several arrays.  This is a **named tuple** meaning that
    each array can be accessed using the following attribute names:

        * ``im``: The binary image of the void space
        * ``dt``: The distance transform of the image
        * ``peaks``: The peaks of the distance transform after applying the
        steps of the SNOW algorithm
        * ``regions``: The void space partitioned into pores using a marker
        based watershed with the peaks found by the SNOW algorithm

    References
    ----------
    [1] Gostick, J. "A versatile and efficient network extraction algorithm
    using marker-based watershed segmenation".  Physical Review E. (2017)

    """
    tup = namedtuple('results', field_names=['im', 'dt', 'peaks', 'regions'])
    im = im.squeeze()
    print('_'*60)
    print("Beginning SNOW Algorithm")

    if im.dtype == 'bool':
        print('Peforming Distance Transform')
        dt = spim.distance_transform_edt(input=im)
    else:
        dt = im
        im = dt > 0

    tup.im = im
    tup.dt = dt

    if sigma > 0:
        print('Applying Gaussian blur with sigma =', str(sigma))
        dt = spim.gaussian_filter(input=dt, sigma=sigma)

    peaks = find_peaks(dt=dt)
    print('Initial number of peaks: ', spim.label(peaks)[1])
    peaks = trim_saddle_points(peaks=peaks, dt=dt, max_iters=500)
    print('Peaks after trimming saddle points: ', spim.label(peaks)[1])
    peaks = trim_nearby_peaks(peaks=peaks, dt=dt)
    peaks, N = spim.label(peaks)
    print('Peaks after trimming nearby peaks: ', N)
    tup.peaks = peaks
    regions = partition_pore_space(im=dt, peaks=peaks)
    regions = randomize_colors(regions)
    tup.regions = regions
    return tup


def find_peaks(dt, r=4, footprint=None):
    r"""
    Returns all local maxima in the distance transform

    Parameters
    ----------
    dt : ND-array
        The distance transform of the pore space.  This may be calculated and
        filtered using any means desired.

    r : scalar
        The size of the structuring element used in the maximum filter.  This
        controls the localness of any maxima. The default is 3 voxels.

    footprint : ND-array
        Specifies the shape of the structuring element used to define the
        neighborhood when looking for peaks.  If none is specified then a
        spherical shape is used (or circular in 2D).

    Returns
    -------
    An ND-array of booleans with ``True`` values at the location of any local
    maxima.

    Notes
    -----
    It is also possible ot the ``peak_local_max`` function from the
    ``skimage.feature`` module as follows:

    ``peaks = peak_local_max(image=dt, min_distance=r, exclude_border=0,
    indices=False)``

    This automatically uses a square structuring element which is significantly
    faster than using a circular or spherical element.
    """
    dt = dt.squeeze()
    im = dt > 0
    if footprint is None:
        if im.ndim == 2:
            footprint = disk
        elif im.ndim == 3:
            footprint = ball
        else:
            raise Exception("only 2-d and 3-d images are supported")
    mx = spim.maximum_filter(dt + 2*(~im), footprint=footprint(r))
    peaks = (dt == mx)*im
    return peaks


def reduce_peaks_to_points(peaks):
    r"""
    Any peaks that are broad or elongated are replaced with a single voxel
    that is located at the center of mass of the original voxels.

    Parameters
    ----------
    peaks : ND-image
        An image containing True values indicating peaks in the distance
        transform

    Returns
    -------
    An array with the same number of isolated peaks as the original image, but
    fewer total voxels.

    Notes
    -----
    The center of mass of a group of voxels is used as the new single voxel, so
    if the group has an odd shape (like a horse shoe), the new voxel may *not*
    lie on top of the original set.
    """
    if peaks.ndim == 2:
        strel = square
    else:
        strel = cube
    markers, N = spim.label(input=peaks, structure=strel(3))
    inds = spim.measurements.center_of_mass(input=peaks,
                                            labels=markers,
                                            index=sp.arange(1, N))
    inds = sp.floor(inds).astype(int)
    # Centroid may not be on old pixel, so create a new peaks image
    peaks = sp.zeros_like(peaks, dtype=bool)
    peaks[tuple(inds.T)] = True
    return peaks


def trim_saddle_points(peaks, dt, max_iters=10):
    r"""
    Removes peaks that were mistakenly identified because they lied on a
    saddle or ridge in the distance transform that was not actually a true
    local peak.

    Parameters
    ----------
    peaks : ND-array
        A boolean image containing True values to mark peaks in the distance
        transform (``dt``)

    dt : ND-array
        The distance transform of the pore space for which the true peaks are
        sought.

    max_iters : int
        The maximum number of iterations to run while eroding the saddle
        points.  The default is 10, which is usually not reached; however,
        a warning is issued if the loop ends prior to removing all saddle
        points.

    Returns
    -------
    An image with fewer peaks than was received.
    """
    if dt.ndim == 2:
        from skimage.morphology import square as cube
    else:
        from skimage.morphology import cube
    labels, N = spim.label(peaks)
    slices = spim.find_objects(labels)
    for i in range(N):
        s = extend_slice(s=slices[i], shape=peaks.shape, pad=10)
        peaks_i = labels[s] == i+1
        dt_i = dt[s]
        im_i = dt_i > 0
        iters = 0
        peaks_dil = sp.copy(peaks_i)
        while iters < max_iters:
            iters += 1
            peaks_dil = spim.binary_dilation(input=peaks_dil,
                                             structure=cube(3))
            peaks_max = peaks_dil*sp.amax(dt_i*peaks_dil)
            peaks_extended = (peaks_max == dt_i)*im_i
            if sp.all(peaks_extended == peaks_i):
                break  # Found a true peak
            elif sp.sum(peaks_extended*peaks_i) == 0:
                peaks_i = False
                break  # Found a saddle point
        peaks[s] = peaks_i
        if iters >= max_iters:
            print('Maximum number of iterations reached, consider' +
                  'running again with a larger value of max_iters')
    return peaks


def trim_nearby_peaks(peaks, dt):
    r"""
    Finds pairs of peaks that are nearer to each other than to the solid phase,
    and removes the peak that is closer to the solid.

    Parameters
    ----------
    peaks : ND-array
        A boolean image containing True values to mark peaks in the distance
        transform (``dt``)

    dt : ND-array
        The distance transform of the pore space for which the true peaks are
        sought.

    Returns
    -------
    An array the same size as ``peaks`` containing a subset of the peaks in
    the original image.

    Notes
    -----
    Each pair of peaks is considered simultaneously, so for a triplet of peaks
    each pair is considered.  This ensures that only the single peak that is
    furthest from the solid is kept.  No iteration is required.
    """
    if dt.ndim == 2:
        from skimage.morphology import square as cube
    else:
        from skimage.morphology import cube
    peaks, N = spim.label(peaks, structure=cube(3))
    crds = spim.measurements.center_of_mass(peaks, labels=peaks,
                                            index=sp.arange(1, N+1))
    crds = sp.vstack(crds).astype(int)  # Convert to numpy array of ints
    # Get distance between each peak as a distance map
    tree = sptl.cKDTree(data=crds)
    temp = tree.query(x=crds, k=2)
    nearest_neighbor = temp[1][:, 1]
    dist_to_neighbor = temp[0][:, 1]
    del temp, tree  # Free-up memory
    dist_to_solid = dt[list(crds.T)]  # Get distance to solid for each peak
    hits = sp.where(dist_to_neighbor < dist_to_solid)[0]
    # Drop peak that is closer to the solid than it's neighbor
    drop_peaks = []
    for peak in hits:
        if dist_to_solid[peak] < dist_to_solid[nearest_neighbor[peak]]:
            drop_peaks.append(peak)
        else:
            drop_peaks.append(nearest_neighbor[peak])
    drop_peaks = sp.unique(drop_peaks)
    # Remove peaks from image
    slices = spim.find_objects(input=peaks)
    for s in drop_peaks:
        peaks[slices[s]] = 0
    return (peaks > 0)



import scipy as sp
from porespy.filters import porosimetry
import matplotlib.pyplot as plt
from collections import namedtuple


class Porosimetry(object):
    r"""
    Simulates a porosimetry experiment on a binary image.  This function is
    equivalent to the morphological image opening and/or the full
    morphology approaches.

    Parameters
    ----------
    im : ND-array
        The Boolean image of the porous material, with True values
        indicating the pore space

    """

    def __init__(self, im, voxel_size=1,):
        self.im = im
        self.voxel_size = voxel_size

    def _get_result(self):
        if not hasattr(self, '_result'):
            print('Simulation has not been run, wait while result is computed')
            self.run()
        return self._result

    result = property(fget=_get_result)

    def run(self, npts=25, sizes=None, inlets=None, access_limited=True):
        temp = porosimetry(im=self.im, npts=npts, sizes=sizes, inlets=inlets,
                           access_limited=access_limited)
        self._result = temp
        return temp

    run.__doc__ = porosimetry.__doc__

    def plot_drainage_curve(self, fig=None, **kwargs):
        r"""
        """
        d = self.get_drainage_data()
        if fig is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        else:
            try:
                ax = fig.get_axes()[0]
            finally:
                ax = fig.get_axes()
        ax.plot(d.radius, d.saturation, **kwargs)
        return fig

    def plot_size_histogram(self, fig=None, **kwargs):
        r"""
        """
        if fig is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        else:
            ax = fig.get_axes()[0]
        counts, bins, bars = ax.hist(self.result.flatten(),
                                     bins=sp.unique(self.result)[1:])
        plt.xlabel('Radius of Invading Sphere')
        plt.ylabel('Number of Invaded Voxels')
        return fig, counts, bins, bars

    def get_drainage_data(self):
        r"""
        Calculate drainage curve based on the image produced by the
        ``porosimetry`` function.

        Returns
        -------
        Rp, Snwp: Two arrays containing (a) the radius of the penetrating
        sphere (in voxels) and (b) the volume fraction of pore phase voxels
        that are accessible from the specfied inlets.

        Notes
        -----
        This function normalizes the invading phase saturation by total pore
        volume of the dry image, which is assumed to be all voxels with a value
        equal to 1.  To do porosimetry on images with large outer regions,
        use the ```find_outer_region``` function then set these regions to 0 in
        the input image.  In future, this function could be adapted to apply
        this check by default.

        """
        im = self.result
        sizes = sp.unique(im)
        R = []
        Snwp = []
        Vp = sp.sum(im > 0)
        for r in sizes[1:]:
            R.append(r)
            Snwp.append(sp.sum(im >= r))
        Snwp = [s/Vp for s in Snwp]
        data = namedtuple('xy_data', ('radius', 'saturation'))
        return data(R, Snwp)



r'''
============
Binarization
============

This module contains functions for thresholding greyscale images into binary
images (i.e. True and False) indicating the phases.

.. autofunction:: simple_otsu

'''

from .__funcs__ import simple_otsu



import numpy as np
import matplotlib.pyplot as plt
from porespy.metrics import porosity
from numba import jit
from mpl_toolkits.mplot3d import Axes3D
from collections import namedtuple


class RandomWalk:
    r"""
    This class generates objects for performing random walk simulations
    on.

    Parameters
    ----------
    im: ndarray of bool
        2D or 3D image
    walkers: int (default = 1000)
        The number of walks to initially perform
    max_steps: int (default = 5000)
        The number of steps to attempt per walk
    stride: int (default = 10)
        The number of steps taken between each coordinate in path_data
    st_frac: float (default = 0.2)
        This fraction of the image (in the centre) is searched for a valid
        starting point for each walk

    Examples
    --------

    Creating a RandomWalk object:

    >>> import porespy as ps
    >>> im = ps.generators.blobs(300)
    >>> rw = ps.simulations.RandomWalk(im, walkers=500, max_steps=3000)

    """

    @jit
    def walk(self, start_point, max_steps, stride=1):
        r"""
        This function performs a single random walk through porous image. It
        returns an array containing the walker path in the image, and the
        walker path in free space.

        Parameters
        ----------
        start_point: array_like
           A sequence containing the indices of a valid starting point. If
           the image is 2D, the first coordinate needs to be zero
        max_steps: int
            The number of steps to attempt per walk
        stride: int
            Number of steps taken between each returned coordinate

        Returns
        --------
        paths: namedtuple
            A tuple containing 2 arrays showing the paths taken by the walker:
            path.pore_space and path.free_space

        Notes
        ------
        This function only performs a single walk through the image. Use
        RandomWalk.-perform_walks() to add more
        """

        ndim = self._ndim
        im = self.im
        if ndim == 3:
            directions = 6
        elif ndim == 2:
            directions = 4
        (z, y, x) = start_point
        if not im[z, y, x]:
            raise ValueError('invalid starting point: not a pore')
        z_max, y_max, x_max = self._shape
        z_free, y_free, x_free = z, y, x
        coords = np.ones((max_steps+1, 3), dtype=int) * (-1)
        free_coords = np.ones((max_steps+1, 3), dtype=int) * (-1)
        coords[0, :] = [z, y, x]
        free_coords[0, :] = [z_free, y_free, x_free]
        steps = 0
        # begin walk
        for step in range(1, max_steps+1):
            x_step, y_step, z_step = 0, 0, 0
            direction = np.random.randint(0, directions)
            if direction == 0:
                x_step += 1
            elif direction == 1:
                x_step -= 1
            elif direction == 2:
                y_step += 1
            elif direction == 3:
                y_step -= 1
            elif direction == 4:
                z_step += 1
            elif direction == 5:
                z_step -= 1
            # checks to make sure image does not go out of bounds
            if x_free+x_step < 0 or y_free+y_step < 0 or z_free+z_step < 0:
                break
            elif (x_free+x_step >= x_max or y_free+y_step >= y_max or
                    z_free+z_step >= z_max):
                break
            else:
                x_free += x_step
                y_free += y_step
                z_free += z_step
            if x+x_step < 0 or y+y_step < 0 or z+z_step < 0:
                break
            elif x+x_step >= x_max or y+y_step >= y_max or z+z_step >= z_max:
                break
            # checks if the step leads to a pore in image
            elif im[z+z_step, y+y_step, x+x_step]:
                x += x_step
                y += y_step
                z += z_step
            steps += 1
            coords[step] = [z, y, x]
            free_coords[step] = [z_free, y_free, x_free]
        path = namedtuple('path', ('pore_space', 'free_space'))
        paths = path(coords[::stride, :], free_coords[::stride, :])
        return paths

    def find_start_point(self, start_frac):
        r"""
        Finds a random valid start point in a porous image, searching in the
        given fraction of the image

        Parameters
        ----------
        start_frac: float
            A value between 0 and 1. Determines what fraction of the image is
            randomly searched for a starting point

        Returns
        --------
        start_point: tuple
            A tuple containing the index of a valid start point.
            If the image is 2D, start_point will have 0 as the first coord
        """

        x_r = self._x_len*start_frac
        y_r = self._y_len*start_frac
        z_r = self._z_len*start_frac
        i = 0
        while True:
            i += 1
            if i > 10000:
                print("failed to find starting point")
                return None
            x = int(self._x_len/2 - x_r/2 + np.random.randint(x_r+1))
            y = int(self._y_len/2 - y_r/2 + np.random.randint(y_r+1))
            z = int(self._z_len/2 - z_r/2 + np.random.randint(z_r+1))
            if self.im[z, y, x]:
                break
        start_point = (z, y, x)
        return start_point

    def perform_walks(self, walkers):
        r"""
        Adds the specified number of walkers to the path_data attribute

        Parameters
        ----------
        walkers: int
            The number of walks to perform. This many rows will be added
            to each of the path_data arrays (path_data.pore_space and
            path_data.free_space)

        Notes
        ------
        This function will continue to use the max_steps and stride value
        stored on the object
        """
        path_data = -1*np.ones((3, walkers, self._max_steps//self._stride+1))
        free_path_data = -1*np.ones((3, walkers,
                                     self._max_steps//self._stride+1))
        paths = namedtuple('path', ('pore_space', 'free_space'))
        for w in range(walkers):
            p = self.find_start_point(self._start_frac)
            path, free_path = self.walk(p, self._max_steps, self._stride)
            path_data[:, w, :] = np.swapaxes(path, 0, 1)
            free_path_data[:, w, :] = np.swapaxes(free_path, 0, 1)
        if self._path_data is None:
            self._path_data = paths(path_data, free_path_data)
        else:
            new_data = np.concatenate((self._path_data.pore_space, path_data),
                                      1)
            new_free_data = np.concatenate((self._path_data.free_space,
                                            free_path_data), 1)
            self._path_data = paths(new_data, new_free_data)
        self._sd_updated = False
        self._sterr_updated = False

    def __init__(self, im, walkers=1000, max_steps=5000, stride=10,
                 start_frac=0.2):
        self.im = np.array(im, ndmin=3)
        self._max_steps = max_steps
        self._stride = stride
        self._start_frac = start_frac
        self.porosity = porosity(im)
        self._ndim = np.ndim(im)
        if self._ndim == 3:
            self._z_len = np.size(im, 0)
            self._y_len = np.size(im, 1)
            self._x_len = np.size(im, 2)
        elif self._ndim == 2:
            self._z_len = 1
            self._y_len = np.size(im, 0)
            self._x_len = np.size(im, 1)
        else:
            raise ValueError('image needs to be 2 or 3 dimensional')
        self._shape = (self._z_len, self._y_len, self._x_len)
        self._path_data = None
        self._sd_data = None
        self._sd_updated = False
        self._sterr_data = None
        self._sterr_updated = False
        self.perform_walks(walkers)

    @property
    def steps(self):
        return np.arange(0, self.max_steps+1, self.stride)

    @property
    def path_data(self):
        return self._path_data

    @property
    def sd_data(self):
        if not self._sd_updated:
            self._get_sq_displacement()
        return self._sd_data

    @property
    def sterr_data(self):
        if not self._sterr_updated:
            self._get_sterr()
        return self._sterr_data

    @property
    def ndim(self):
        return self._ndim

    @property
    def shape(self):
        return self._shape

    @property
    def max_steps(self):
        return self._max_steps

    @property
    def total_walkers(self):
        return np.size(self._path_data.pore_space, 1)

    @property
    def stride(self):
        return self._stride

    @property
    def start_frac(self):
        return self._start_frac

    def _get_sq_displacement(self):
        r"""
        Generates squared displacement arrays for pore_space and free_space
        """
        squared_displacement = namedtuple('square_displacement',
                                          ('pore_space', 'free_space'))
        paths = self._path_data.pore_space
        free_paths = self._path_data.free_space
        sd = np.ones(paths.shape)*-1
        sd_free = np.ones(free_paths.shape)*-1
        for w in range(np.size(paths, 1)):
            path = paths[:, w, :]
            path = path[:, path[0, :] >= 0]
            free_path = free_paths[:, w, :]
            free_path = free_path[:, free_path[0, :] >= 0]
            for i in range(np.size(path, 1)):
                sd[:, w, i] = (path[:, i]-path[:, 0])**2
                sd_free[:, w, i] = (free_path[:, i] - free_path[:, 0])**2
        self._sd_data = squared_displacement(sd, sd_free)
        self._sd_updated = True

    def _get_sterr(self):
        r"""
        """
        standard_error = namedtuple('standard_error',
                                    ('pore_space', 'free_space'))
        sd, sd_f = self.sd_data
        ste = np.zeros((4, np.size(self.steps)))
        ste_f = np.zeros((4, np.size(self.steps)))
        for col in range(np.size(sd, 2)):
            ste[3, col] = np.std(np.sum(sd[:, np.where(sd[0, :, col] >= 0),
                                 col], 0))
            ste_f[3, col] = np.std(np.sum(sd_f[:,
                                   np.where(sd_f[0, :, col] >= 0), col], 0))
            ste[3, col] /= np.sqrt(np.size(np.where(sd[0, :, col] >= 0)))
            ste_f[3, col] /= np.sqrt(np.size(np.where(sd[0, :, col] >= 0)))
        for d in range(3):
            for col in range(np.size(sd, 2)):
                ste[d, col] = np.std(sd[d, np.where(sd[d, :, col] >= 0), col])
                ste_f[d, col] = np.std(sd_f[d, np.where(sd_f[d, :, col] >= 0),
                                       col])
                ste[d, col] /= np.sqrt(np.size(np.where(sd[0, :, col] >= 0)))
                ste_f[d, col] /= np.sqrt(np.size(np.where(sd[0, :, col] >= 0)))
        self._sterr_data = standard_error(ste, ste_f)
        self._sterr_updated = True

    def get_msd(self, direction=None):
        r"""
        """
        d = direction
        sd, sd_f = self.sd_data
        msd = np.zeros(np.size(self.steps))
        msd_f = np.zeros(np.size(self.steps))
        mean_sd = namedtuple('mean_squared_displacement',
                             ('pore_space', 'free_space'))
        if d is None:
            for col in range(np.size(sd, 2)):
                msd[col] = np.mean(np.sum(sd[:, np.where(sd[0, :, col] >= 0),
                                   col], 0))
                msd_f[col] = np.mean(np.sum(sd_f[:,
                                     np.where(sd_f[0, :, col] >= 0), col], 0))
        else:
            for col in range(np.size(sd, 2)):
                msd[col] = np.mean(sd[d, np.where(sd[0, :, col] >= 0), col])
                msd_f[col] = np.mean(sd_f[d, np.where(sd_f[0, :, col] >= 0),
                                     col])
        return mean_sd(msd, msd_f)

    def show_msd(self, step_range=None, direction=None, showstderr=True):
        r"""
        Graphs mean squared displacement in pore space and free space
        vs. number of steps taken

        Parameters
        -----------
        step_range: array_like (optional)
            A sequence with the min and max range of steps to plot. If no
            argument is given, the range will be zero to max_steps
        direction: int (optional)
            The axis along which to show mean squared displacement in. If no
            argument is given, total msd is shown
        showstderr: bool (optional)
            If set to True, the graph will show error bars

        Returns
        --------
        out: namedtuple
        """
        if step_range is None:
            step_range = (0, self.max_steps)
        start, stop = step_range
        d = direction
        start = start//self.stride
        stop = stop//self.stride
        sd, sd_f = self.sd_data
        ste, ste_f = self.sterr_data
        steps = self.steps
        msd, msd_f = self.get_msd(direction)
        slopes = namedtuple('slope', ('pore_space', 'free_space'))
        if d is None:
            ste = ste[3, :]
            ste_f = ste_f[3, :]
        else:
            ste = ste[d, :]
            ste_f = ste_f[d, :]
        p = np.polyfit(steps[start:stop], msd[start:stop], 1)
        p_f = np.polyfit(steps[start:stop], msd_f[start:stop], 1)
        if showstderr:
            plt.errorbar(steps[start:stop], msd[start:stop],
                         yerr=ste[start:stop])
            plt.errorbar(steps[start:stop], msd_f[start:stop],
                         yerr=ste_f[start:stop])
        else:
            plt.plot(steps[start:stop], msd[start:stop])
            plt.plot(steps[start:stop], msd_f[start:stop])
        return slopes(p[0], p_f[0])

    def _get_path(self, walker):
        r"""
        Returns matplotlib figures for show_path and save_path functions
        """
        z, y, x = self.shape
        path = self.path_data.pore_space[:, walker, :]
        path = path[:, path[0, :] >= 0]
        free_path = self.path_data.free_space[:, walker, :]
        free_path = free_path[:, free_path[0, :] >= 0]
        max_i = np.size(path, 1) - 1
        if self._ndim == 3:
            size = (7*x/y, 7*z/y)
            fig_im = plt.figure(num=1, figsize=size)
            fig_im = Axes3D(fig_im)
            fig_im.plot(path[2, :], path[1, :], path[0, :], 'c')
            fig_im.plot([path[2, 0]], [path[1, 0]], [path[0, 0]], 'g.')
            fig_im.plot([path[2, max_i]], [path[1, max_i]], [path[0, max_i]],
                        'r.')
            fig_im.set_xlim3d(0, x)
            fig_im.set_ylim3d(0, y)
            fig_im.set_zlim3d(0, z)
            fig_im.invert_yaxis()
            plt.title('Path in Pore Space')
            fig_free = plt.figure(num=2, figsize=size)
            fig_free = Axes3D(fig_free)
            fig_free.plot(free_path[2, :], free_path[1, :], free_path[0, :],
                          'c')
            fig_free.plot([free_path[2, 0]], [free_path[1, 0]],
                          [free_path[0, 0]], 'g.')
            fig_free.plot([free_path[2, max_i]], [free_path[1, max_i]],
                          [free_path[0, max_i]], 'r.')
            fig_free.set_xlim3d(0, x)
            fig_free.set_ylim3d(0, y)
            fig_free.set_zlim3d(0, z)
            fig_free.invert_yaxis()
            plt.title('Path in Free Space')
        elif self._ndim == 2:
            size = (5, 5)
            fig_im = plt.figure(num=1, figsize=size)
            plt.plot(path[2, :], path[1, :], 'c')
            plt.plot(path[2, 0], path[1, 0], 'g.')
            plt.plot(path[2, max_i], path[1, max_i], 'r.')
            plt.xlim((0, x))
            plt.ylim((0, y))
            plt.gca().invert_yaxis()
            plt.title('Path in Pore Space')
            fig_free = plt.figure(num=2, figsize=size)
            plt.plot(free_path[2, :], free_path[1, :], 'c')
            plt.plot(free_path[2, 0], free_path[1, 0], 'g.')
            plt.plot(free_path[2, max_i], free_path[1, max_i], 'r.')
            plt.xlim((0, x))
            plt.ylim((0, y))
            plt.gca().invert_yaxis()
            plt.title('Path in Free Space')
        return(fig_im, fig_free)

    def show_path(self, walker):
        r"""
        Shows the selected walkers path through pore space and free space
        using a matplotlib figure

        Parameters
        -----------
        walker: int
            The walker whose path will be shown
        """
        self._get_path(walker)
        plt.show()

    def save_path(self, walker, f_name):
        r"""
        Saves the selected walkers path through pore space and free space

        Parameters
        -----------
        walker: int
            The walker whose path will be saved

        f_name: string
            A path to a filename. Note that this function will add the suffix
            pore_space and free_space to differentiate between paths
        """

        fig_im, fig_free = self._get_path(walker)
        plt.savefig(f_name+'free_space.png')
        plt.figure(num=1)
        plt.savefig(f_name+'pore_space.png')



r"""
=======
Metrics
=======

A suite of tools for determine key metrics about an image.  Typically these are
applied to an image after applying a filter, but a few functions can be applied
directly to the binary image.

.. autofunction:: chord_length_counts
.. autofunction:: chord_length_distribution
.. autofunction:: pore_size_density
.. autofunction:: pore_size_distribution
.. autofunction:: porosity
.. autofunction:: representative_elementary_volume
.. autofunction:: two_point_correlation_bf
.. autofunction:: two_point_correlation_fft

"""

from .__funcs__ import chord_length_counts
from .__funcs__ import chord_length_distribution
from .__funcs__ import pore_size_density
from .__funcs__ import pore_size_distribution
from .__funcs__ import porosity
from .__funcs__ import representative_elementary_volume
from .__funcs__ import two_point_correlation_bf
from .__funcs__ import two_point_correlation_fft



from skimage import filters
import porespy as ps
import matplotlib.pyplot as plt
import scipy.ndimage as spim
import scipy as sp
from skimage.morphology import ball, disk, square, cube
from skimage.segmentation import clear_border


def simple_otsu(im, trim_solid=True):
    r"""
    Uses Otsu's method to find a threshold, then uses binary opening and
    closing to remove noise.

    Parameters
    ----------
    im : ND-image
        The greyscale image of the porous medium to be binarized.

    trim_solid : Boolean
        If True (default) then all solid voxels not connected to an image
        boundary are trimmed.

    Returns
    -------
    An ND-image the same size as ``im`` but with True and False values
    indicating the binarized or segmented phases.

    Examples
    --------

    >>> im = ps.generators.blobs([300, 300], porosity=0.5)
    >>> im = ps.generators.add_noise(im)
    >>> im = spim.gaussian_filter(im, sigma=1)
    >>> im = simple_otsu(im)

    """
    if im.ndim == 2:
        ball = disk
        cube = square
    im = sp.pad(array=im, pad_width=1, mode='constant', constant_values=1)
    val = filters.threshold_otsu(im)
    im = im >= val
    # Remove speckled noise from void and solid phases
    im = spim.binary_closing(input=im, structure=ball(1))
    im = spim.binary_opening(input=im, structure=ball(1))
    # Clean up edges
    im = spim.binary_closing(input=im, structure=cube(3))
    im = spim.binary_opening(input=im, structure=cube(3))
    im = im[[slice(1, im.shape[d]-1) for d in range(im.ndim)]]
    temp = clear_border(~im)
    im = im + temp
    return im



# ***********************************************************************************
# Copyright 2010 - 2016 Paulo A. Herrera. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2.Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY PAULO A. HERRERA ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING
# BUT NOT LIMITED TO, PROCUREMEN OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ***********************************************************************************

import struct
import numpy as np
import sys

# Map numpy dtype to struct format
np_to_struct = {'int8': 'b',
                'uint8': 'B',
                'int16': 'h',
                'uint16': 'H',
                'int32': 'i',
                'uint32': 'I',
                'int64': 'q',
                'uint64': 'Q',
                'float32': 'f',
                'float64': 'd'}


def _get_byte_order_char():
    # Check format in https://docs.python.org/3.5/library/struct.html
    if sys.byteorder == "little":
        return '<'
    else:
        return '>'

# ================================
#        Python interface
# ================================


def writeBlockSize(stream, block_size):
    fmt = _get_byte_order_char() + 'Q'  # Write size as unsigned long
    #  long == 64 bits unsigned integer
    stream.write(struct.pack(fmt, block_size))


def writeArrayToFile(stream, data):
    #  stream.flush() # this should not be necessary
    assert (data.ndim == 1 or data.ndim == 3)
    fmt = _get_byte_order_char() + str(data.size)
    fmt = fmt + np_to_struct[data.dtype.name]  # > for big endian

    # Check if array is contiguous
    assert (data.flags['C_CONTIGUOUS'] or data.flags['F_CONTIGUOUS'])

    # NOTE: VTK expects data in FORTRAN order
    # This is only needed when a multidimensional array has C-layout
    dd = np.ravel(data, order='F')

    bin = struct.pack(fmt, *dd)
    stream.write(bin)

# ==============================================================================


def writeArraysToFile(stream, x, y, z):
    # Check if arrays have same shape and data type
    assert(x.size == y.size == z.size), "Different array sizes."
    assert(x.dtype.itemsize ==
           y.dtype.itemsize == z.dtype.itemsize), "Different item sizes."

    nitems = x.size
    itemsize = x.dtype.itemsize

    fmt = _get_byte_order_char() + str(1) + np_to_struct[x.dtype.name]
    # > for big endian

    # Check if arrays are contiguous
    assert (x.flags['C_CONTIGUOUS'] or x.flags['F_CONTIGUOUS'])
    assert (y.flags['C_CONTIGUOUS'] or y.flags['F_CONTIGUOUS'])
    assert (z.flags['C_CONTIGUOUS'] or z.flags['F_CONTIGUOUS'])

    # NOTE: VTK expects data in FORTRAN order
    # This is only needed when a multidimensional array has C-layout
    xx = np.ravel(x, order='F')
    yy = np.ravel(y, order='F')
    zz = np.ravel(z, order='F')

    # eliminate this loop by creating a composed array.
    for i in range(nitems):
        bx = struct.pack(fmt, xx[i])
        by = struct.pack(fmt, yy[i])
        bz = struct.pack(fmt, zz[i])
        stream.write(bx)
        stream.write(by)
        stream.write(bz)



r'''
=======
PoreSpy
=======

PoreSpy is a collection of functions that are especially useful for analyzing
binary images of porous materials, typically produced by X-ray tomography.
The functions in this collection are often simple combinations of other
standard image analysis functions, so really only offer the convenience of
organizing the functions into one place, and sparing you the trouble of working
them out.

-------
Modules
-------

This package consists of several modules, the purposes of which are given below:

+------------------------+----------------------------------------------------+
| **binarization**       | Convert greyscale images to binary                 |
+------------------------+----------------------------------------------------+
| **filters**            | Process images based on structural features        |
+------------------------+----------------------------------------------------+
| **generators**         | Make artificial images for testing & illustration  |
+------------------------+----------------------------------------------------+
| **metrics**            | Extract information from images                    |
+------------------------+----------------------------------------------------+
| **network_extraction** | Extract pore network models from images            |
+------------------------+----------------------------------------------------+
| **simulations**        | Performing complex simulations directly on an image|
+------------------------+----------------------------------------------------+
| **tools**              | Utilities for altering & manipulating images       |
+------------------------+----------------------------------------------------+
| **visualization**      | Quickly and easily visualiz 3D images              |
+------------------------+----------------------------------------------------+
| **io**                 | Import and export image data in various formats    |
+------------------------+----------------------------------------------------+

-------------
Example Usage
-------------

>>> import porespy as ps
>>> import matplotlib.pyplot as plt
>>> im = ps.generators.blobs([100, 100])
>>> mip = ps.filters.porosimetry(im)
>>> PcSw = ps.metrics.pore_size_distribution(mip)
>>> fig = plt.plot(*PcSw)

----------------
Related Packages
----------------

PoreSpy relies heavily on two general image analysis packages:
**scipy.ndimage** and **scikit-image** also known as **skimage**.  The former
contains an assortment of veyr general image analysis tools such as image
morphology filters, while the latter offers more complex but still general
functions such as watershed segmentation.  PoreSpy does not duplicate any of
these general functions so you will also have to install and learn how to
use them to get the most from PoreSpy.  The functions in PoreSpy are generally
built up using several of the more general functions offered by **skimage**
and **scipy**.  There are a few functions in PoreSpy that are implimented
natively, but only when necessary.

-----------
Image Types
-----------

PoreSpy is meant to work on single-channel, binary images.  Such images are
conveniently represented by Numpy arrays, hence all references to an *image* is
equivalent to an *array*.  It is further assumed that the arrays are binarized,
meaning 1's or True values indicating the void space, and 0's or False values
for the solid.

-----------
Limitations
-----------
Although *scikit-image* and *scipy.ndimage* have a wide assortment of
functions, they are not always the fastest implementation.  It is often faster
to use ImageJ for many things, such as distance transforms and image
morphology.  The advantage of PoreSpy is the flexibility offered by the
Python environment.

'''

__version__ = "0.3.9"

from . import tools
from . import network_extraction
from . import visualization
from . import simulations
from . import metrics
from . import generators
from . import binarization
from . import filters
from . import io
